
  ANF STATIC CHECKER
  ==================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: ANFChecker.lhs,v 1.1 2003/03/31 09:11:12 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  The "check" function verifies the well-formedness of an ANF program.
  Well-formedness is a pretty weak condition, basically enforcing the expected
  scoping rules for the program. The function returns "GOOD" on success, or
  "BAD msgs" on error, where msgs is a list of error messages describing the
  problems with the program.

> module ANFChecker (
>  Verdict(..), check
> ) where

> import List
> import Set
> import FiniteMap
> import Diag
> import Posn
> import ANFPrims
> import ANFSyntax

  First, we need a type for the return value:

> data Verdict = GOOD | BAD Msgs

> instance Eq Verdict where
>  (==) GOOD GOOD = True
>  (==) (BAD _) (BAD _) = True
>  (==) _ _ = False

> instance Show Verdict where
>  showsPrec _ GOOD = \s -> "the program is well-formed" ++ s
>  showsPrec p (BAD msgs) = showsPrec p msgs

  Verdicts may be combined using the "&&&" operator:

> (&&&) :: Verdict -> Verdict -> Verdict
> GOOD &&& GOOD = GOOD
> GOOD &&& bad@(BAD _) = bad
> bad@(BAD _) &&& GOOD = bad
> (BAD m1) &&& (BAD m2) = BAD (m1 ++ m2)

  The environment "Gamma" is simply a set of all variables currently in scope:

> type Gamma = Set Xp

  Let's call the set of primitives "prims":

> prims :: Gamma
> prims = fst (foldl (enumerate) (emptySet, -1) primNames)
>  where
>   enumerate (gamma, gen) name = (addToSet gamma (IDp gen name posn), gen - 1)
>   posn = startPosn "<prelude>"

  Now we can define the "check" function for the ANF programs:

> check :: Ep -> Verdict
> check = checkE prims

  "check" is defined in terms of an expression checker:

> checkE :: Gamma -> Ep -> Verdict

  We need a separate rule for each ANF construct:

> checkE gamma (RETp vv) =
>  checkVV gamma vv
> checkE gamma (RETAp v vv) =
>  checkV gamma v &&&
>  checkVV gamma vv
> checkE gamma (LETp x v e _) =
>  let
>   gamma' = addToSet gamma x
>  in
>   checkV gamma v &&&
>   checkE gamma' e
> checkE gamma (LETAp xv v vv e _) =
>  let
>   gamma' = foldl (addToSet) gamma xv
>  in
>   checkUniqXV xv &&&
>   checkV gamma v &&&
>   checkVV gamma vv &&&
>   checkE gamma' e
> checkE gamma (LETRECp fv e _) =
>  let
>   gamma' = foldl (addToSet) gamma (map (\ (FUNp x xv p) -> x) fv)
>  in
>   checkUniqFV fv &&&
>   checkFV gamma' fv &&&
>   checkE gamma' e
> checkE gamma (IFp v e1 e2 _) =
>  checkV gamma v &&&
>  checkE gamma e1 &&& checkE gamma e2
> checkE gamma (SYNTAX_ERROR msgs) = BAD msgs

> checkF :: Gamma -> Fp -> Verdict
> checkF gamma (FUNp x xv e) =
>  let
>   gamma' = foldl (addToSet) gamma xv
>  in
>   checkUniqXV xv &&&
>   checkE gamma' e

> checkV :: Gamma -> Vp -> Verdict
> checkV gamma (VARp x) =
>  if x `elementOf` gamma then
>   GOOD
>  else
>   BAD [parseError (exprPosn x) ("Undefined variable <" ++ show x ++ ">")]
> checkV gamma (INTp _ _) = GOOD

> checkFV :: Gamma -> FVp -> Verdict
> checkFV gamma fv = foldl (&&&) GOOD (map (checkF gamma) fv)

> checkVV :: Gamma -> VVp -> Verdict
> checkVV gamma vv = foldl (&&&) GOOD (map (checkV gamma) vv)

  Finally, the "checkUniq*" functions verify that all identifiers
  in a list of entities being defined simultaneously are unique:

> checkUniqFV :: FVp -> Verdict
> checkUniqFV fv = checkUniqFV' fv fv

> checkUniqFV' :: FVp -> FVp -> Verdict
> checkUniqFV' (f@(FUNp x xv e):fv) fv' =
>  let
>   Just f' = find (\ (FUNp x' xv' e') -> x' == x) fv'
>   verdict' = checkUniqFV' fv fv'
>  in
>   if exprPosn f' == exprPosn f then
>    verdict'
>   else
>    BAD [parseError (exprPosn f) (msg (exprPosn f'))] &&& verdict'
>  where
>   msg posn = "Duplicate definition of function <" ++ show x ++ ">\n" ++
>    "(previous definition found on line " ++
>    show (lineno posn) ++ " in " ++ show (filename posn) ++ ".)"
> checkUniqFV' [] fv' = GOOD

> checkUniqXV :: XVp -> Verdict
> checkUniqXV xv = checkUniqXV' xv xv

> checkUniqXV' :: XVp -> XVp -> Verdict
> checkUniqXV' (x:xv) xv' =
>  let
>   Just x' = find (== x) xv'
>   verdict' = checkUniqXV' xv xv'
>  in
>   if exprPosn x' == exprPosn x then
>    verdict'
>   else
>    BAD [parseError (exprPosn x) msg] &&& verdict'
>  where
>   msg = "Duplicate definition of parameter <" ++ show x ++ ">"
> checkUniqXV' [] xv' = GOOD
