
  ANF INTERPRETER
  ===============

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: ANFInterp.lhs,v 1.5 2003/03/31 09:09:38 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  This module provides a rudimentary interpreter for ANF programs, which
  defines the transitional semantics for our ANF language. The program must
  be syntactically and semantically well-formed, that is, it must conform
  to the well-formedness condition defined by "check" in "ANFChecker". The
  result of evaluation is a runtime value, which may be "RUNTIME_ERROR"
  if a condition that is not cought by the "check" (such as a division by
  zero) is encountered during evaluation.

> module ANFInterp (
>  Vr, eval
> ) where

> import FiniteMap
> import ANFPrims
> import ANFSyntax

  We must definie a type for runtime values, which includes integer constants,
  closures, and a handful of predefined primitives:

> data Vr = INTr Integer |
>  CL XV E |
>  PRIM (VVr -> VVr) |
>  RUNTIME_ERROR String
> type VVr = [Vr]

> instance Show Vr where
>
>  showsPrec _ (INTr x) = shows x
>  showsPrec _ (CL xv e) = \ s -> "<function>" ++ s
>  showsPrec _ (PRIM op) = \ s -> "<primitive>" ++ s
>  showsPrec _ (RUNTIME_ERROR msg) = \ s -> "<" ++ msg ++ ">" ++ s
>
>  showList (vr:vvr) = shows vr . showList' vvr
>   where
>    showList' [] = id
>    showList' vvr = (\ s -> ", " ++ s) . showList vvr
>  showList [] = id

  We also need a type for continuations, which include activation records,
  and the usual "stop" continuation:

> data K = STOP | AR XV E Gamma K

  The ANF environment consists of mapping from identifiers to values:

> type Gamma = FiniteMap Int Vr

  "vextend" extends an environment by a single variable binding:

> vextend :: Gamma -> X -> Vr -> Gamma
> vextend gamma (ID id) vr = addToFM gamma id vr

  "vvextend" extends an environment by a list of variables:

> vvextend :: Gamma -> XV -> VVr -> Maybe Gamma
> vvextend gamma [] [] = Just gamma
> vvextend gamma (x:xv) (vr:vvr) = vvextend (vextend gamma x vr) xv vvr
> vvextend gamma _ _ = Nothing

  "fextend" extends the environment by a list of "letrec" bindings:

> fextend :: Gamma -> FV -> Gamma
> fextend gamma fv =
>  foldl (\ gamma (FUN x xv e) -> vextend gamma x (CL xv e)) gamma fv

  The "vlookup" function lookup up a function in the environment:

> vlookup :: Gamma -> V -> Vr
> vlookup gamma (VAR x@(ID id)) =
>  case lookupFM gamma id of
>   Just vr -> vr
>   Nothing -> error ("undefined variable <" ++ show x ++ ">")
> vlookup gamma (INT c)  = INTr c

  And one for looking up a list of function parameters:

> vvlookup :: Gamma -> VV -> VVr
> vvlookup gamma vv = map (vlookup gamma) vv

  The initial environment consists of the list of primitives:

> primOps :: FiniteMap String (VVr -> VVr)
> primOps = listToFM [
>  ("add", \vvr ->
>	case vvr of
>	 [INTr x, INTr y] -> [INTr (x + y)]
>	 _ -> [RUNTIME_ERROR
>		"runtime error: \"add\" called with incorrect parameters"]),
>  ("sub", \vvr ->
>	case vvr of
>	 [INTr x, INTr y] -> [INTr (x - y)]
>	 _ -> [RUNTIME_ERROR
>		"runtime error: \"sub\" called with incorrect parameters"]),
>  ("mul", \vvr ->
>	case vvr of
>	 [INTr x, INTr y] -> [INTr (x * y)]
>	 _ -> [RUNTIME_ERROR
>		"runtime error: \"mul\" called with incorrect parameters"]),
>  ("div", \vvr ->
>	case vvr of
>	 [INTr x, INTr y] ->
>	  case y of
>	   0 -> [INTr (div x y)]
>	   _ -> [RUNTIME_ERROR "runtime error: division by zero"]
>	 _ -> [RUNTIME_ERROR 
>		"runtime error: \"div\" called with incorrect parameters"]),
>  ("rem", \vvr ->
>	case vvr of
>	 [INTr x, INTr y] ->
>	  case y of
>	   0 -> [INTr (rem x y)]
>	   _ -> [RUNTIME_ERROR "runtime error: division by zero"]
>	 _ -> [RUNTIME_ERROR
>		"runtime error: \"rem\" called with incorrect parameters"])]

> primIds :: FiniteMap String Int
> primIds = fst (foldl (enumerate) (emptyFM, -1) primNames)
>  where enumerate (ids, gen) name = (addToFM ids name gen, gen - 1)

> prims :: Gamma
> prims =
>  if keysFM primOps /= keysFM primIds then
>   error "list of primitives in ANFInterp incompatible with ANFPrims"
>  else
>   foldl (mkPrim) emptyFM (keysFM primOps)
>  where
>   mkPrim gamma name =
>    let
>     Just id = lookupFM primIds name
>     Just op = lookupFM primOps name
>    in
>     addToFM gamma id (PRIM op)

  Now we can define the evaluation functions for the ANF programs:

> eval :: E -> VVr
> eval e = evalE prims STOP e

  The program evaluation function is defined in term of an expression
  evaluation function:

> evalE :: Gamma -> K -> E -> VVr

  We need a separate transition rule for each ANF construct:

> evalE gamma k (RET vv) =
>  let
>   vvr = vvlookup gamma vv
>  in
>   case k of
>    AR xv e' gamma' k' ->
>     case vvextend gamma' xv vvr of
>      Just gamma'' -> evalE gamma'' k' e'
>      Nothing -> [RUNTIME_ERROR "invalid arity"]
>    STOP -> vvr

> evalE gamma k (RETA v vv) =
>  let
>   vr = vlookup gamma v
>   vvr = vvlookup gamma vv
>  in
>   case vr of
>    CL xv' e' ->
>     case vvextend gamma xv' vvr of
>      Just gamma'' -> evalE gamma'' k e'
>      Nothing -> [RUNTIME_ERROR "invalid arity"]
>    PRIM op ->
>     case k of
>      AR xv e' gamma' k' ->
>       case vvextend gamma' xv (op vvr) of
>        Just gamma'' -> evalE gamma'' k' e'
>        Nothing -> [RUNTIME_ERROR "invalid arity"]
>      STOP -> op vvr
>    _ -> [RUNTIME_ERROR "invalid function object (must be a closure)"]

> evalE gamma k (LET x v e) =
>  let
>   vr = vlookup gamma v
>  in
>   evalE (vextend gamma x vr) k e

> evalE gamma k (LETA xv v vv e) =
>  let
>   vr = vlookup gamma v
>   vvr = vvlookup gamma vv
>  in
>   case vr of
>    CL xv' e' ->
>     case vvextend gamma xv' vvr of
>      Just gamma'' -> evalE gamma'' (AR xv e gamma k) e'
>      Nothing -> [RUNTIME_ERROR "invalid arity"]
>    PRIM op ->
>     case vvextend gamma xv (op vvr) of
>      Just gamma'' -> evalE gamma'' k e
>      Nothing -> [RUNTIME_ERROR "invalid arity"]
>    _ -> [RUNTIME_ERROR "invalid function object (must be a closure)"]

> evalE gamma k (LETREC fv e) =
>  evalE (fextend gamma fv) k e

> evalE gamma k (IF v e1 e2) =
>  let
>   vr = vlookup gamma v
>  in
>   case vr of
>    INTr 0 -> evalE gamma k e2
>    INTr _ -> evalE gamma k e1
>    _ -> [RUNTIME_ERROR "invalid condition parameter (must be an integer)"]
