
  ANF RENAMER
  ===========

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: ANFRenamer.lhs,v 1.1 2003/03/31 09:11:12 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  The renamer assigns globally-unique identifiers to all variables, which
  greatly simplifies future program analysis. The original identifier
  information may be discarded when the debug information is stripped by
  "strip"; the new identifiers will be printed as "@n", where "n" is the
  globally-unique variable number.

  By convention, we number identifiers with positive numbers, reserving 0
  for free (undefined) variables and negative numbers for primitives.

  The "rename" function returns the new expression tree.

> module ANFRenamer (
>  rename
> ) where

> import FiniteMap
> import Posn
> import ANFSyntax
> import ANFPrims

  First, we need an environment, which maps variable names to identifier
  numbers:

> type Gamma = FiniteMap String Int

  The initial environment contains mappings for all primitives:

> prims :: Gamma
> prims = fst (foldl (enumerate) (emptyFM, -1) primNames)
>  where enumerate (gamma, gen) name = (addToFM gamma name gen, gen - 1)

  Now, we define functions for extending the environment and assigning
  new identifier numbers to variable names:

> introduceF :: Int -> Gamma -> Fp -> (Gamma, Int)
> introduceF gen gamma (FUNp x xv e) = introduceX gen gamma x

> introduceX :: Int -> Gamma -> Xp -> (Gamma, Int)
> introduceX gen gamma (IDp id name _) = (addToFM gamma name gen, gen + 1)

  The vector versions of the above are a little tricky. When the list contains
  duplicate definitions of an identifier, the last definition in the list will
  replace the previous ones in the resulting "gamma". This is precisely the
  effect we want, since it makes the error visible to the type checker, which
  operates on the identifier numbers rather than names.

> introduceFV :: Int -> Gamma -> FVp -> (Gamma, Int)
> introduceFV gen gamma fv = foldl (introduce) (gamma, gen) fv
>  where introduce (gamma, gen) f = introduceF gen gamma f

> introduceXV :: Int -> Gamma -> XVp -> (Gamma, Int)
> introduceXV gen gamma xv = foldl (introduce) (gamma, gen) xv
>  where introduce (gamma, gen) x = introduceX gen gamma x

  Finally, we are ready to define "rename":

> rename :: Ep -> Ep
> rename e = fst (renameE 1 prims e)

> renameE :: Int -> Gamma -> Ep -> (Ep, Int)
> renameE gen gamma (RETp vv) =
>  let
>   (vv') = renameVV gamma vv
>  in
>   (RETp vv', gen)
> renameE gen gamma (RETAp v vv) =
>  let
>   (v' ) = renameV  gamma v
>   (vv') = renameVV gamma vv
>  in
>   (RETAp v' vv', gen)
> renameE gen gamma (LETp x v e posn) =
>  let
>   (gamma', gen1) = introduceX gen  gamma  x
>   (x'          ) = renameX         gamma' x
>   (v'          ) = renameV         gamma  v
>   (e',     gen2) = renameE    gen1 gamma' e
>  in
>   (LETp x' v' e' posn, gen2)
> renameE gen gamma (LETAp xv v vv e posn) =
>  let
>   (gamma', gen1) = introduceXV gen  gamma  xv
>   (xv'         ) = renameXV         gamma' xv
>   (v'          ) = renameV          gamma  v
>   (vv'         ) = renameVV         gamma  vv
>   (e',     gen2) = renameE     gen1 gamma' e
>  in
>   (LETAp xv' v' vv' e' posn, gen2)
> renameE gen gamma (LETRECp fv e posn) =
>  let
>   (gamma', gen1) = introduceFV gen  gamma  fv
>   (fv',    gen2) = renameFV    gen1 gamma' fv
>   (e',     gen3) = renameE     gen2 gamma' e
>  in
>   (LETRECp fv' e' posn, gen3)
> renameE gen gamma (IFp v e1 e2 posn) =
>  let
>   (v'       ) = renameV      gamma v
>   (e1', gen1) = renameE gen  gamma e1
>   (e2', gen2) = renameE gen1 gamma e2
>  in
>   (IFp v' e1' e2' posn, gen2)
> renameE gen gamma e@(SYNTAX_ERROR msg) = (e, gen)

> renameF :: Int -> Gamma -> Fp -> (Fp, Int)
> renameF gen gamma (FUNp x xv e) =
>  let
>   (gamma', gen1) = introduceXV gen  gamma  xv
>   (x'          ) = renameX          gamma' x
>   (xv'         ) = renameXV         gamma' xv
>   (e',     gen2) = renameE     gen1 gamma' e
>  in
>   (FUNp x' xv' e', gen2)

> renameV :: Gamma -> Vp -> Vp
> renameV gamma (VARp x) = VARp (renameX gamma x)
> renameV gamma v@(INTp _ _) = v

> renameX :: Gamma -> Xp -> Xp
> renameX gamma (IDp _ name posn) =
>  IDp (lookupWithDefaultFM gamma 0 name) name posn

> renameFV :: Int -> Gamma -> FVp -> (FVp, Int)
> renameFV gen gamma (f:fv) =
>  let
>   (f',  gen1) = renameF  gen  gamma f
>   (fv', gen2) = renameFV gen1 gamma fv
>  in
>   (f':fv', gen2)
> renameFV gen gamma [] = ([], gen)

> renameVV :: Gamma -> VVp -> VVp
> renameVV gamma vv = map (renameV gamma) vv

> renameXV :: Gamma -> XVp -> XVp
> renameXV gamma xv = map (renameX gamma) xv
