
  ABSTRACT SYNTAX FOR ANF PROGRAMS
  ================================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: ANFSyntax.lhs,v 1.5 2003/03/31 09:09:39 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  This module defines the abstract syntax for ANF programs as a set of Haskell
  types. Initially, all syntactic elements are annotated with their position
  within the input stream ("Ep" and friends.) Once the program has been
  type-checked, the position information may be discarded, and, since
  processing the annonated expressions is somewhat annoying, we define the
  second set of data structures without position annotations ("E" and
  friends.) The module also defines a pretty-printer for unannotated ANF
  expressions by instantiating the "Show" class for all AST types.

> module ANFSyntax (
>  E(..), F(..), V(..), X(..),
>	FV, VV, XV,
>  Ep(..), Fp(..), Vp(..), Xp(..),
>	FVp, VVp, XVp,
>  ANFElem(..), strip
> ) where

> import Posn
> import Diag
> import ANFPrims

  The AST for ANF programs:

> data E  = RET VV |
> 	    RETA V VV |
>	    LET X V E |
>	    LETA XV V VV E |
>	    LETREC FV E |
>	    IF V E E
> data F  = FUN X XV E
> data V  = VAR X | INT Integer
> data X  = ID Int
> type FV = [F]
> type VV = [V]
> type XV = [X]

> instance Eq X where
>  (==) (ID x) (ID y) = (x == y)

> instance Ord X where
>  compare (ID x) (ID y) = compare x y

  The AST annotated with position information:

> data Ep  = RETp VVp |
> 	     RETAp Vp VVp |
>	     LETp Xp Vp Ep Posn |
>	     LETAp XVp Vp VVp Ep Posn |
>	     LETRECp FVp Ep Posn |
>	     IFp Vp Ep Ep Posn |
>            SYNTAX_ERROR Msgs
> data Fp  = FUNp Xp XVp Ep
> data Vp  = VARp Xp | INTp Integer Posn
> data Xp  = IDp Int String Posn
> type FVp = [Fp]
> type VVp = [Vp]
> type XVp = [Xp]

> instance Eq Xp where
>  (==) (IDp x _ _) (IDp y _ _) = (x == y)

> instance Ord Xp where
>  compare (IDp x _ _) (IDp y _ _) = compare x y

  Convinient access to the position information:

> class ANFElem a where
>  exprPosn :: a -> Posn

> instance ANFElem Ep where
>  exprPosn (RETp (v:vv)) = exprPosn v
>  exprPosn (RETAp v vv) = exprPosn v
>  exprPosn (LETp x v e posn) = posn
>  exprPosn (LETAp xv v vv e posn) = posn
>  exprPosn (LETRECp fv e posn) = posn
>  exprPosn (IFp x e1 e2 posn) = posn
>  exprPosn (SYNTAX_ERROR msgs) = undefined
> instance ANFElem Fp where
>  exprPosn (FUNp x xv e) = exprPosn x
> instance ANFElem Vp where
>  exprPosn (VARp x) = exprPosn x
>  exprPosn (INTp x posn) = posn
> instance ANFElem Xp where
>  exprPosn (IDp x name posn) = posn

  A function which removes position informatoin from the program:

> strip :: Ep -> E
> strip = stripE

> stripE :: Ep -> E
> stripE (RETp vv) = RET (stripVV vv)
> stripE (RETAp v vv) = RETA (stripV v) (stripVV vv)
> stripE (LETp x v e _) = LET (stripX x) (stripV v) (stripE e)
> stripE (LETAp xv v vv e _) =
>  LETA (stripXV xv) (stripV v) (stripVV vv) (stripE e)
> stripE (LETRECp fv e _) = LETREC (stripFV fv) (stripE e)
> stripE (IFp v e1 e2 _) = IF (stripV v) (stripE e1) (stripE e2)
> stripE (SYNTAX_ERROR _) = error "cannot strip an invalid program"

> stripF :: Fp -> F
> stripF (FUNp x xv e) = FUN (stripX x) (stripXV xv) (stripE e)

> stripV :: Vp -> V
> stripV (VARp x) = VAR (stripX x)
> stripV (INTp x _) = INT x

> stripX :: Xp -> X
> stripX (IDp x _ _) = ID x

> stripFV :: FVp -> FV
> stripFV = map (stripF)

> stripVV :: VVp -> VV
> stripVV = map (stripV)

> stripXV :: XVp -> XV
> stripXV = map (stripX)

  The pretty-printer for stripped expressions:

> instance Show E where
>  showsPrec _ = showE 0
> instance Show V where
>  showsPrec _ = showV
>  showList = showVV
> instance Show F where
>  showsPrec _ = showF 0
>  showList = showFV 0
> instance Show X where
>  showsPrec _ = showX
>  showList = showXV

> showE :: Int -> E -> ShowS
> showE i (RET vv) = showVV vv
> showE i (RETA v vv) = showV v . showC '(' . showVV vv . showC ')'
> showE i (LET x v e) = showS "let\n" . showLET i x v e
> showE i (LETA xv v vv e) = showS "let\n" . showLETA i xv v vv e
> showE i (LETREC [] e) =
>  showS "letrec in\n" .
>  indent (i + 1) . showE (i + 1) e
> showE i (LETREC fv e) =
>  showS "letrec\n" .
>  showFV (i + 1) fv . showC '\n' .
>  indent i . showS "in\n" .
>  indent (i + 1) . showE (i + 1) e
> showE i (IF v e1 e2) =
>  showS "if " . showV v . showS " then\n" .
>  indent (i + 1) . showE (i + 1) e1 . showC '\n' .
>  indent i . showS "else\n" .
>  indent (i + 1) . showE (i + 1) e2

> showLET :: Int -> X -> V -> E -> ShowS
> showLET i x v e =
>  indent (i + 1) . showX x . showS " = " . showV v . showC '\n' .
>  showIN i e

> showLETA :: Int -> XV -> V -> VV -> E -> ShowS
> showLETA i xv v vv e =
>  indent (i + 1) .
>  showXV xv . showS " = " . showV v . showC '(' . showVV vv . showS ")\n" .
>  showIN i e

> showIN :: Int -> E -> ShowS
> showIN i (LET x v e) = showLET i x v e
> showIN i (LETA xv v vv e) = showLETA i xv v vv e
> showIN i e = 
>  indent i . showS "in\n" .
>  indent (i + 1) . showE (i + 1) e

> showF :: Int -> F -> ShowS
> showF i (FUN x xv e) =
>  showX x . showC '(' . showXV xv . showS ") =\n" .
>  indent (i + 1) . showE (i + 1) e

> showV :: V -> ShowS
> showV (VAR x) = showX x
> showV (INT x) = shows x

> showX :: X -> ShowS
> showX (ID id) =
>  if id < 0 then
>   if -id <= length primNames then
>    showS (primNames!!(-id - 1))
>   else
>    showS "@@" . shows (-id)
>  else
>   showC '@' . shows id

> showFV :: Int -> FV -> ShowS
> showFV i (f:fv) = indent i . showF i f . showFV' i fv
>  where
>   showFV' i [] = id
>   showFV' i fv = showC '\n' . showFV i fv
> showFV i [] = id

> showVV :: VV -> ShowS
> showVV = showA

> showXV :: XV -> ShowS
> showXV = showA

  Now the variant for annotated expression:

> instance Show Ep where
>  showsPrec _ = showEp 0
> instance Show Vp where
>  showsPrec _ = showVp
>  showList = showVVp
> instance Show Fp where
>  showsPrec _ = showFp 0
>  showList = showFVp 0
> instance Show Xp where
>  showsPrec _ = showXp
>  showList = showXVp

> showEp :: Int -> Ep -> ShowS
> showEp i (RETp vv) = showVVp vv
> showEp i (RETAp v vv) = showVp v . showC '(' . showVVp vv . showC ')'
> showEp i (LETp x v e _) = showS "let\n" . showLETp i x v e
> showEp i (LETAp xv v vv e _) = showS "let\n" . showLETAp i xv v vv e
> showEp i (LETRECp [] e _) =
>  showS "letrec in\n" .
>  indent (i + 1) . showEp (i + 1) e
> showEp i (LETRECp fv e _) =
>  showS "letrec\n" .
>  showFVp (i + 1) fv . showC '\n' .
>  indent i . showS "in\n" .
>  indent (i + 1) . showEp (i + 1) e
> showEp i (IFp v e1 e2 _) =
>  showS "if " . showVp v . showS " then\n" .
>  indent (i + 1) . showEp (i + 1) e1 . showC '\n' .
>  indent i . showS "else\n" .
>  indent (i + 1) . showEp (i + 1) e2
> showEp i (SYNTAX_ERROR msgs) = shows msgs

> showLETp :: Int -> Xp -> Vp -> Ep -> ShowS
> showLETp i x v e =
>  indent (i + 1) . showXp x . showS " = " . showVp v . showC '\n' .
>  showINp i e

> showLETAp :: Int -> XVp -> Vp -> VVp -> Ep -> ShowS
> showLETAp i xv v vv e =
>  indent (i + 1) .
>  showXVp xv . showS " = " . showVp v . showC '(' . showVVp vv . showS ")\n" .
>  showINp i e

> showINp :: Int -> Ep -> ShowS
> showINp i (LETp x v e _) = showLETp i x v e
> showINp i (LETAp xv v vv e _) = showLETAp i xv v vv e
> showINp i e = 
>  indent i . showS "in\n" .
>  indent (i + 1) . showEp (i + 1) e

> showFp :: Int -> Fp -> ShowS
> showFp i (FUNp x xv e) =
>  showXp x . showC '(' . showXVp xv . showS ") =\n" .
>  indent (i + 1) . showEp (i + 1) e

> showVp :: Vp -> ShowS
> showVp (VARp x) = showXp x
> showVp (INTp x _) = shows x

> showXp :: Xp -> ShowS
> showXp (IDp x name _) = showS name

> showFVp :: Int -> FVp -> ShowS
> showFVp i (f:fv) = indent i . showFp i f . showFVp' i fv
>  where
>   showFVp' i [] = id
>   showFVp' i fv = showC '\n' . showFVp i fv
> showFVp i [] = id

> showVVp :: VVp -> ShowS
> showVVp = showA

> showXVp :: XVp -> ShowS
> showXVp = showA

  And some additional helpers for "show":

> showA :: (Show a) => [a] -> ShowS
> showA (a:av) = shows a . showA' av
>  where
>   showA' [] = id
>   showA' av = showS ", " . showA av
> showA [] = id

> showS :: String -> ShowS
> showS (c:s) = showC c . showS s
> showS [] = id

> showC :: Char -> ShowS
> showC c = \ s -> c : s

> indent :: Int -> ShowS
> indent 0 = id
> indent n = showS "  " . indent (n - 1)
