
  DIAGNOSTICS MODULE
  ==================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: Diag.lhs,v 1.1 2003/03/31 09:11:12 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  This module provides a library of interfaces for managing and formatting
  syntactic and semantic error diagnostic messages.  The library is built
  around the "Msg" ADT.  "Msg" objects can be constructed using one of the two
  functions "parseError" and "parseWarning", and messages are stored in a lazy
  list.  At the end of the program, any generated messages are emitted to
  "stderr" using the "emitDiagnostics" function, as described below.

  The library is quite generic and contains many more features that are
  necessary for our purposes.

> module Diag (
>  Msg, Msgs, parseError, parseWarning, emitDiagnostics,
>  fatalError, 
>  noInputFiles,
>  notEnoughCmdlineArgs,
>  tooManyCmdlineArgs,
>  invalidCmdlineFlag,
>  missingCmdlineFlagArg,
>  incompatibleCmdlineFlags,
>  repeatedCmdlineFlag,
> ) where

> import Common
> import Charset (isbreak)
> import IO (hPutStr, hPutStrLn, stderr)
> import Posn (Posn, filename, lineoff, lineno, colno)
> import System (exitFailure, getProgName, getEnv)


  THE MESSAGE TYPE
  ================  

  First, we must define "Msg" as a (severity, posn, text) tuple.

> data Severity = E | W
> data Msg = M Severity Posn String
> type Msgs = [Msg]

> instance Show Severity where
>  showsPrec _ E = \s -> "Error" ++ s
>  showsPrec _ W = \s -> "Warning" ++ s

  Define a pair of simple functions for constructing error and warning
  objects, respectively.

> parseError :: Posn -> String -> Msg
> parseError posn msg = M E posn msg

> parseWarning :: Posn -> String -> Msg
> parseWarning posn msg = M W posn msg


  FIRST, A SIMPLE ERROR VIEWER
  ============================

> instance Show Msg where
>  showsPrec _ = showMessage
>  showList = showMessages

> showMessage :: Msg -> ShowS
> showMessage (M severity posn msg) = \s ->
>  s1 ++ " on line " ++ s2 ++ " in file " ++ s3 ++ ":\n" ++ msg ++ s
>  where
>   s1 = show severity
>   s2 = show (lineno posn)
>   s3 = show (filename posn)

> showMessages [] = id
> showMessages [msg] = showMessage msg
> showMessages (msg:msgs) =
>  showMessage msg . (\s -> '\n' : s ) . showMessages msgs


 THE "emitDiagnostics" FUNCTION
 ==============================

 "emitDiagnostics"  is called  from "main"  to print  any errors  and warnings
 accumulated during processing  of the source files.  If it encounters any
 true errors in the error list, it will abort the program with "exitFailure".

 The  content of  the file,  as  fetched by  "readFile", must  be supplied  to
 "emitDiagnostics"  to  allow  it  to  annotate the  error  message  with  the
 offending line in the file.

 Finally,  "emitDiagnostics" can be  used in  one or  two modes.   In "strict"
 mode, all warnings  are treated as errors. Otherwise,  warnings are displayed
 but are otherwise ignored in the error count.

> emitDiagnostics :: Msgs -> Bool -> String -> IO ()
> emitDiagnostics msgs strict contents  =
>  do
>   abort <- showMsgs msgs strict contents maxErrors
>   if abort then exitFailure else return ()

 "maxErrors" represents  the maximum number  of error messages  printed before
 aborting the program.

> maxErrors :: Int
> maxErrors = 5

 "showMsgs"  prints accumulated error  messages to  stderr.  All  warnings are
 emitted, but at most "maxErrors"  error messages are printed before giving up.
 The function returns a flag indicating if any errors have been encountered.

> showMsgs :: Msgs -> Bool -> String -> Int -> IO Bool
> showMsgs [] _ _ _ = return False
> showMsgs ((M W posn msg):rest) False contents remaining = 
>  do
>   showMsg "Warning" posn msg contents
>   abort <- showMsgs rest False contents remaining
>   return abort
> showMsgs _ _ _ 0 = 
>  do
>   fatalError "Too many errors to continue."
>   return True
> showMsgs (M _ posn msg:rest) strict contents remaining = 
>  do
>   showMsg "Error" posn msg contents
>   showMsgs rest strict contents (remaining - 1)
>   return True

 Next, I define a function that prints a formatted error message. The messages
 are in the format:

	[progname] Error/Warning on line N in "file":
	           The actual error message.
	--------------------------------------------------------------
	the line of code with the error position indicated by "^"
	-----^--------------------------------------------------------

 where the screen width is given by "getColumns".  Error printing is performed
 directly from "main", so "formatMessage"  is supplied the content of the file.

> showMsg :: String -> Posn -> String -> String -> IO ()
> showMsg severity p msg contents =
>  do
>   columns <- getColumns
>   let
>    line0 = severity ++ " on line " ++ show line
>	     ++ " in " ++ show (filename p) ++ ": "
>    line1 = replicate columns '-'
>    line2 = take columns (takeWhile (not.isbreak) (drop (lineoff p) contents))
>    line3
>     | col1 < columns = replicate col1 '-' ++ "^" ++ replicate col2 '-'
>     | otherwise      = line1
>    line = lineno p		-- starts at 1
>    col1 = colno p - 1		-- starts at 0
>    col2 = columns - 1 - col1	-- any columns after the '^'
>    in
>     do
>      putMsg (line0 : lines msg)
>      hPutStrLn stderr line1
>      hPutStrLn stderr line2
>      hPutStrLn stderr line3

 Print the first two  lines of a message. The first line  is prefixed with the
 message prefix, and the remaining lines are indented by an appropriate number
 of spaces.

> putMsg :: Strings -> IO () 
> putMsg (line:rest) =
>  do
>   prefix <- getPrefix
>   hPutStrLn stderr (prefix ++ line)
>   putMsg' (map (\_ -> ' ') prefix) rest

> putMsg' :: String -> Strings -> IO ()
> putMsg' _ [] = return ()
> putMsg' prefix (line:rest) =
>  do
>   hPutStrLn stderr (prefix ++ line)
>   putMsg' prefix rest


 TERMINAL INTERFACE
 ==================

 To  format  the message  as  described  aboce, we  need  to  obtain the  base
 component of our program name and the screen width from the COLUMNS variable,
 if set.

> getPrefix :: IO String
> getPrefix =
>  do
>   progname <- getProgName
>   let pretty = show (basename progname) in
>    return ("[" ++ tail (take (length pretty - 1) pretty) ++ "] ")

> getColumns :: IO Int
> getColumns = catch getColumns' (\err -> return defaultWidth)
>  where
>   getColumns' =
>    do
>     env <- getEnv "COLUMNS"
>     let (columns, rest) = read env in
>      if columns < minWidth \/ rest /= "" then
>       return defaultWidth
>      else
>       return columns
>   minWidth = 30
>   defaultWidth = 80


 FATAL ERRORS
 ============

 In the  remainder of  this module I  define a  bunch of common  fatal errors.
 Each error is an I/O action that formats and prints the supplied messages and
 aborts the program with "exitFailure".

> fatalError :: String -> IO ()
> fatalError msg =
>  do
>   putMsg (lines msg)
>   exitFailure

> noInputFiles :: IO ()
> noInputFiles = fatalError msg
>  where
>   msg =
>    "No input files specified.\n\
>    \Use the -? option to see detailed synopsis."

> notEnoughCmdlineArgs :: Int -> IO ()
> notEnoughCmdlineArgs n = fatalError msg
>  where
>   msg =
>    "Insufficient number of command-line arguments.\n\
>    \(" ++ show n ++ " expected.)\n\
>    \Use the -? option to see detailed synopsis."

> tooManyCmdlineArgs :: Int -> Strings -> IO ()
> tooManyCmdlineArgs n args = fatalError msg
>  where
>   msg =
>    "Unexpected command-line argument " ++ show (args !! n) ++ "\n\
>    \(" ++ expected ++ ")\n\
>    \Use the -? option to see detailed synopsis."
>   expected =
>    if n == 0 then
>     "no arguments expected"
>    else if n == 1 then
>     "only one argument expected"
>    else
>     "only " ++ show n ++ " arguments expected"

> invalidCmdlineFlag :: String -> IO ()
> invalidCmdlineFlag f = fatalError msg
>  where
>   msg =
>    "Invalid command-line flag " ++ f' ++ "\n\
>    \Use the -? flag to see detailed synopsis."
>   f' = '-' : drop 2 (take (length f'' - 1) f'')
>   f'' = show f

> missingCmdlineFlagArg :: String -> IO ()
> missingCmdlineFlagArg f = fatalError msg
>  where
>   msg =
>    "Command-line flag " ++ f ++ " requires an argument.\n\
>    \Use the -? flag to see detailed synopsis."

> incompatibleCmdlineFlags :: String -> String -> IO ()
> incompatibleCmdlineFlags e f = fatalError msg
>  where
>   msg =
>    "Command-line flags " ++ e ++ " and " ++ f ++ " are mutually exclusive.\n\
>    \Use the -? flag to see detailed synopsis."

> repeatedCmdlineFlag :: String -> IO ()
> repeatedCmdlineFlag f = fatalError msg
>  where
>   msg =
>    "Command-line flag " ++ f ++ " cannot be specified more than once.\n\
>    \Use the -? flag to see detailed synopsis."
