
  TOKEN POSITION LIBRARY
  ======================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: Posn.lhs,v 1.1 2003/03/31 09:11:12 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  This small module implements the "Posn" ADT which keeps track of the
  character and token positions in the input stream. The position of each
  token consists of the name of the source file, the logical position (line
  number, column number) which provides easy access to the human-readable
  position of a token, and a stream position (line offset, character offset),
  which allows the diagnostic library from the "Diag" module to quickly find
  the line containing the error.

  The position of the first character in an input stream is obtained using the
  "startPosn" function. Various elements of the position can be obtained using
  the "filename", "lineoff", "charoff", "lineno" and "colno" functions.

  Line and character offsets are counted from 0. Column and line numbers are
  counted from 1. Tab stops are set "tabWidth" columns apart.

  The "annotateString" function annotates all characters in a string with
  token positions, replaces all line breaks with a single '\n' character, and
  appends a single '\r' to the last line in the input. "stripString" performs
  the reverse transformation, except, of course, all line breaks remain
  encoded as a single '\n' character.

> module Posn (
>  Posn, startPosn, tabWidth, filename, lineoff, charoff, lineno, colno,
>  PChar, PString, annotateString, stripString
> ) where

> import Charset

> data Posn = P { filename :: String, lineoff, charoff, lineno, colno :: Int }

> instance Eq Posn where
>  (==) p1 p2 = (charoff p1 == charoff p2)

> instance Ord Posn where
>  compare p1 p2 = compare (charoff p1) (charoff p2)

> type PChar = (Char, Posn)
> type PString = [PChar]

> startPosn :: String -> Posn
> startPosn name =
>  P { filename = name, lineoff = 0, charoff = 0, lineno = 1, colno = 1 }

> tabWidth :: Int
> tabWidth = 8

 A position object can be modified using "adv", "nc", "nl" and "ht" functions,
 as follows:

 "adv n" (advance) advances the position by "n" characters and columns. Note
 that "adv" cannot be used to move to the new line, next tab stop, or to
 backtrack:

> adv :: Int -> Posn -> Posn
> adv n p | n > 0 = p { charoff = charoff p + n,
>		        colno = colno p + n }

 "nc" (next character) advances by one character and column position:

> nc :: Posn -> Posn
> nc = adv 1

 "nl" (new line) advances the position by one character and to the beginning
 of the next line. This is used for both UNIX-style line feeds and 
 MacOS-style carriage returns that are not followed by a line feed:

> nl :: Posn -> Posn
> nl p = p { lineoff = charoff p + 1,
>            charoff = charoff p + 1,
>            lineno = lineno p + 1,
>            colno = 1 }

 "cr" (carriage-return) adjust the position by a line-feed that appears
 immediately after a carriage-return:

> cr :: Posn -> Posn
> cr p = p { lineoff = lineoff p + 1,
>            charoff = charoff p + 1 }

 "ht" (horizontal tab) advances the position by one character and to the
 beginning of the next tab stop.

> ht :: Posn -> Posn
> ht p = p { charoff = charoff p + 1,
>            colno = tabWidth * div (colno p - 1 + tabWidth) tabWidth + 1 }

 The "annotateString" function annotates a string with the position
 of each character. It ensures that the file ends in a newline, and
 converts all line breaks to a line-feed character:

> annotateString :: String -> String -> PString
> annotateString name xs = annotateNL (startPosn name) xs

> annotate :: Posn -> String -> PString
> annotate p [] = ('\n', p) : ('\r', p) : []
> annotate p ('\r':xs) = ('\n', p) : annotateCR (nl p) xs
> annotate p ('\t':xs) = ('\t', p) : annotate (ht p) xs
> annotate p (c:xs)
>  | isbreak c = ('\n', p) : annotateNL (nl p) xs
>  | otherwise = (c, p) : annotate (nc p) xs

> annotateNL :: Posn -> String -> PString
> annotateNL p [] = ('\r', p) : []
> annotateNL p ('\r':xs) = ('\n', p) : annotateCR (nl p) xs
> annotateNL p ('\t':xs) = ('\t', p) : annotate (ht p) xs
> annotateNL p (c:xs)
>  | isbreak c = ('\n', p) : annotateNL (nl p) xs
>  | otherwise = (c, p) : annotate (nc p) xs

> annotateCR :: Posn -> String -> PString
> annotateCR p [] = ('\r', p) : []
> annotateCR p ('\n':xs) = annotateNL (cr p) xs
> annotateCR p ('\r':xs) = ('\n', p) : annotateCR (nl p) xs
> annotateCR p ('\t':xs) = ('\t', p) : annotate (ht p) xs
> annotateCR p (c:xs)
>  | isbreak c = ('\n', p) : annotateNL (nl p) xs
>  | otherwise = (c, p) : annotate (nc p) xs

 "stripString" removes the annotations introduced by "annotateString":

> stripString :: PString -> String
> stripString s = take (length s - 1) (map (\ (c,p) -> c) s)
