
  SSA STATIC CHECKER
  ==================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: SSAChecker.lhs,v 1.1 2003/03/31 09:11:12 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  The "check" function verifies the well-formedness of an SSA program.
  Well-formedness is a pretty weak condition, basically enforcing the expected
  scoping rules and the dominance property for the program. The function
  returns "GOOD" on success, or "BAD msgs" on error, where msgs is a list of
  error messages describing the problems with the program.

> module SSAChecker (
>  Verdict (..),  check
> ) where

> import List
> import Set
> import FiniteMap
> import Posn
> import Diag
> import SSAPrims
> import SSASyntax

  First, we need a type for the return value:

> data Verdict = GOOD | BAD Msgs

> instance Eq Verdict where
>  (==) GOOD GOOD = True
>  (==) (BAD _) (BAD _) = True
>  (==) _ _ = False

> instance Show Verdict where
>  showsPrec _ GOOD = \s -> "the program is well-formed" ++ s
>  showsPrec p (BAD msgs) = showsPrec p msgs

  Verdicts may be combined using the "&&&" operator:

> (&&&) :: Verdict -> Verdict -> Verdict
> GOOD &&& GOOD = GOOD
> GOOD &&& bad@(BAD _) = bad
> bad@(BAD _) &&& GOOD = bad
> (BAD m1) &&& (BAD m2) = BAD (m1 ++ m2)

  Now, we need a few environments: "Gamma", which contains the set of all
  variables defined in the current scope and "Lambda", which represents a
  mapping from labels to block signatures, which describe the contexts in
  which a jump to the block well-formed:

> type Gamma = Set Xp
> type Lambda = FiniteMap Xp BlockSig

  We must nail down the concept of a "block signature". It is simply a finite
  map of ``come from labels to environments containing variables that must
  be defined in the block jumping to the label. The domain of the map defines
  all blocks which may jump to the label.

> type BlockSig = FiniteMap Lp Gamma

  To compute the signature of a block, we process all "PHI" nodes in the block,
  adding labels and variables found in the "GV" lists to the signature:

> blockSig :: Bp -> BlockSig
> blockSig b = blockSig' emptyFM (blockExpr_p b)

> blockSig' :: BlockSig -> Ep -> BlockSig
> blockSig' sig (PHIp x gv e) = blockSig' (foldl (extend) sig gv) e
>  where
>   extend sig g =
>    let
>     l = cameFrom g
>     gamma = lookupWithDefaultFM sig emptySet l
>    in
>     case g of
>      FROMp l v ->
>       case v of
>        (VARp x) -> addToFM sig l (addToSet gamma x)
>        v -> addToFM sig l gamma
>      BAD_FROMp l x _ -> addToFM sig l (addToSet gamma x)
> blockSig' sig (LETp x v e) =  blockSig' sig e
> blockSig' sig (LETAp x v vv e) = blockSig' sig e
> blockSig' sig (GOTOp x _) = sig
> blockSig' sig (RETp vv _) = sig
> blockSig' sig (RETAp v vv _) = sig
> blockSig' sig (IFp v e1 e2 _) = blockSig' (blockSig' sig e1) e2

  Let's now define another auxiliary function for extracting the label from
  a "phi" parameter:

> cameFrom :: Gp -> Lp
> cameFrom (FROMp l v) = l
> cameFrom (BAD_FROMp l x _) = l

  The initial environment contains mappings for all primitives:

> prims :: Gamma
> prims = fst (foldl (enumerate) (emptySet, -1) primNames)
>  where
>   enumerate (gamma, gen) name = (addToSet gamma (IDp gen name posn), gen - 1)
>   posn = startPosn "<prelude>"

  Armed with these data types, we can define the checker function for an
  SSA program. The initial "Gamma" contains the set of primitives for the
  program:

> check :: Pp -> Verdict
> check p = checkP prims p

> checkP :: Gamma -> Pp -> Verdict
> checkP gamma (PROGp fv e) =
>  let
>   gamma' = foldl (addToSet) gamma (map (\ (PROCp x xv e bv _) -> x) fv)
>  in
>   checkUniqFV fv &&&
>   checkFV gamma' fv &&&
>   checkE gamma' emptyFM (STARTp (exprPosn e)) e
> checkP gamma (SYNTAX_ERROR msgs) = BAD msgs

> checkF :: Gamma -> Fp -> Verdict
> checkF gamma (PROCp x xv e bv posn) =
>  let
>   gamma' = foldl (addToSet) gamma xv
>   lambda' = foldl (extend) emptyFM bv
>  in
>   checkUniqBV bv &&&
>   checkE gamma' lambda' (STARTp posn) e &&&
>   checkBV gamma' lambda' bv
>  where
>   extend lambda b = addToFM lambda (blockLabel_p b) (blockSig b)

> checkB :: Gamma -> Lambda -> Bp -> Verdict
> checkB gamma lambda (BLOCKp x e) =
>  checkE gamma lambda (LABELp x) e
> checkB gamma lambda (GROUPp x e bv) =
>  let
>   lambda' = foldl (extend) lambda bv
>  in
>   checkE gamma lambda (LABELp x) e &&&
>   checkBV gamma lambda bv
>  where
>   extend lambda b = addToFM lambda (blockLabel_p b) (blockSig b)

> checkE :: Gamma -> Lambda -> Lp -> Ep -> Verdict
> checkE gamma lambda l (PHIp x gv e) =
>  let
>   gamma' = addToSet gamma x
>  in 
>   checkAmbiguousGV gv &&&
>   checkUniqGV gv &&&
>   checkGV lambda l gv (exprPosn x) &&&
>   checkE gamma' lambda l e
> checkE gamma lambda l (LETp x v e) =
>  let
>   gamma' = addToSet gamma x
>  in
>   checkV gamma v &&&
>   checkE gamma' lambda l e
> checkE gamma lambda l (LETAp xv v vv e) =
>  let
>   gamma' = foldl (addToSet) gamma xv
>  in
>   checkUniqXV xv &&&
>   checkV gamma v &&&
>   checkVV gamma vv &&&
>   checkE gamma' lambda l e
> checkE gamma lambda l (GOTOp x posn) =
>  case lookupFM lambda x of
>   Just sig ->
>    if isEmptyFM sig then
>     GOOD
>    else
>     case lookupFM sig l of
>      Just gamma' ->
>       let
>        missing = gamma' `minusSet` gamma
>       in
>        if isEmptySet missing then
>         GOOD
>        else
>         BAD [parseError posn (msg1 ++ show (setToList missing))]
>      Nothing -> BAD [parseError posn msg2]
>   Nothing -> BAD [parseError posn msg3]
>  where
>   msg1 = "Undefined variables required by <" ++ show x ++ ">:\n"
>   msg2 = "Label <" ++ show x ++ "> is not reachable from <" ++ show l ++ ">"
>   msg3 = "Undefined label <" ++ show x ++ ">"
> checkE gamma lambda l (RETp vv _) =
>  checkVV gamma vv
> checkE gamma lambda l (RETAp v vv _) =
>  checkV gamma v &&&
>  checkVV gamma vv
> checkE gamma lambda l (IFp v e1 e2 _) =
>  checkV gamma v &&&
>  checkE gamma lambda l e1 &&&
>  checkE gamma lambda l e2

> checkV :: Gamma -> Vp -> Verdict
> checkV env (VARp x) =
>  if x `elementOf` env then
>   GOOD
>  else
>   BAD [parseError (exprPosn x) msg]
>  where
>   msg = "Undefined variable <" ++ show x ++ ">"
> checkV env (INTp _ _) = GOOD

> checkFV :: Gamma -> FVp -> Verdict
> checkFV gamma fv = foldl (&&&) GOOD (map (checkF gamma) fv)

> checkBV :: Gamma -> Lambda -> BVp -> Verdict
> checkBV gamma lambda bv = foldl (&&&) GOOD (map (checkB gamma lambda) bv)

> checkGV :: Lambda -> Lp -> GVp -> Posn -> Verdict
> checkGV lambda (STARTp _) gv posn =
>  BAD [parseError posn msg]
>  where msg = "\"phi\" expressions are not allowed in an entry block"
> checkGV lambda (LABELp x) gv posn =
>  let
>   Just sig = lookupFM lambda x
>   required = mkSet (keysFM sig)
>   given = mkSet (map (\ g -> cameFrom g) gv)
>   missing = required `minusSet` given
>  in
>   if isEmptySet missing then
>    GOOD
>   else
>    BAD [parseError posn (msg ++ show (setToList missing))]
>  where
>   msg = "Missing source labels in the \"phi\" expression:\n"

> checkVV :: Gamma -> VVp -> Verdict
> checkVV gamma vv = foldl (&&&) GOOD (map (checkV gamma) vv)

  Finally, the "checkUniq*" functions verify that all identifiers
  in a list of entities being defined simultaneously are unique:

> checkUniqFV :: FVp -> Verdict
> checkUniqFV fv = checkUniqFV' fv fv

> checkUniqFV' :: FVp -> FVp -> Verdict
> checkUniqFV' (f@(PROCp x xv e bv _):fv) fv' =
>  let
>   Just f' = find (\ (PROCp x' xv' e' bv' _) -> (x' == x)) fv'
>   verdict' = checkUniqFV' fv fv'
>  in
>   if exprPosn f' == exprPosn f then
>    verdict'
>   else
>    BAD [parseError (exprPosn f) (msg (exprPosn f'))] &&& verdict'
>  where
>   msg posn = "Duplicate definition of procedure <" ++ show x ++ ">\n" ++
>    "(previous definition found on line " ++
>    show (lineno posn) ++ " in " ++ show (filename posn) ++ ".)"
> checkUniqFV' [] fv' = GOOD

> checkUniqBV :: BVp -> Verdict
> checkUniqBV bv =
>  let
>   xv = collect [] bv
>  in
>   checkUniqBV' xv xv
>  where
>   collect xv (BLOCKp x e:bv') = x : collect xv bv'
>   collect xv (GROUPp x e bv:bv') = x : collect (collect xv bv') bv
>   collect xv [] = xv

> checkUniqBV' :: XVp -> XVp -> Verdict
> checkUniqBV' (x:xv) xv' =
>  let
>   Just x' = find (== x) xv'
>   verdict' = checkUniqBV' xv xv'
>  in
>   if exprPosn x' == exprPosn x then
>    verdict'
>   else
>    BAD [parseError (exprPosn x) (msg (exprPosn x'))] &&& verdict'
>  where
>   msg posn = "Duplicate definition of label <" ++ show x ++ ">\n" ++
>    "(previous definition found on line " ++
>    show (lineno posn) ++ " in " ++ show (filename posn) ++ ".)"
> checkUniqBV' [] bv' = GOOD

> checkUniqGV :: GVp -> Verdict
> checkUniqGV gv = checkUniqGV' gv gv

> checkUniqGV' :: GVp -> GVp -> Verdict
> checkUniqGV' (g:gv) gv' =
>  let
>   Just g' = find (\ g' -> cameFrom g' == l) gv'
>   verdict' = checkUniqGV' gv gv'
>  in
>   if exprPosn g' == exprPosn g then
>    verdict'
>   else
>    BAD [parseError (exprPosn g) msg] &&& verdict'
>  where
>   l = cameFrom g
>   msg = "Duplicate definition of \"phi\" parameter for <" ++ show l ++ ">"
> checkUniqGV' [] gv' = GOOD

> checkUniqXV :: XVp -> Verdict
> checkUniqXV xv = checkUniqXV' xv xv

> checkUniqXV' :: XVp -> XVp -> Verdict
> checkUniqXV' (x:xv) xv' =
>  let
>   Just x' = find (== x) xv'
>   verdict' = checkUniqXV' xv xv'
>  in
>   if exprPosn x' == exprPosn x then
>    verdict'
>   else
>    BAD [parseError (exprPosn x) msg] &&& verdict'
>  where
>   msg = "Duplicate definition of parameter <" ++ show x ++ ">"
> checkUniqXV' [] xv' = GOOD

> checkAmbiguousGV :: GVp -> Verdict
> checkAmbiguousGV gv = foldl (&&&) GOOD (map (checkAmbiguousG) gv)

> checkAmbiguousG :: Gp -> Verdict
> checkAmbiguousG (FROMp l v) = GOOD
> checkAmbiguousG (BAD_FROMp l x posns) = BAD [parseError (exprPosn x) msg]
>  where
>   (line:lines) = map (lineno) posns
>   msg = "Ambigous reference to variable <" ++ show x ++ ">.\n" ++
>         "This variable is defined in block <" ++ show l ++ "> on lines:\n" ++
>         show line ++ foldl (\s l -> s ++ ", " ++ show l) "" lines ++ "."
