
  THE SSA PROGRAM PARSER
  ======================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: SSAParser.lhs,v 1.5 2003/03/31 09:09:39 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  This module implements a simple recursive-descent parser for our concrete
  SSA syntax. The output is annotated with position information, and should be
  stripped before printing. The expression positions are useful for generating
  useful error messages during parsing and typechecking.

> module SSAParser (
>  parse
> ) where

> import Posn
> import Diag
> import SSASyntax
> import SSAScanner

> parse :: String -> String -> Pp
> parse name content =
>  case parseP (scanTokens name content) of
>   (p, [EOF_TOKEN _]) -> p
>   (_, ERR_TOKEN posn msg:_) -> SYNTAX_ERROR [parseError posn msg]
>   (_, t:_) -> SYNTAX_ERROR [parseError (toknPosn t) msg]
>  where
>   msg = "unexpected tokens after the program entry expression"


	P ::= fv e

> parseP :: Tokens -> (Pp, Tokens)
> parseP t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseP t0 =
>  let
>   (fv, t1) = parseFV t0
>   (e,  t2) = parseE  t1
>  in
>   (PROGp fv e, t2)

	F ::= "proc" X "(" XVopt ")" "{" E BV "}"

> parseF :: Tokens -> (Fp, Tokens)
> parseF t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseF (PROC_TOKEN posn:t1) =
>  let
>   (x,  t2) = parseX      t1
>   (    t3) = parseLPAREN t2
>   (xv, t4) = parseXVopt  t3
>   (    t5) = parseRPAREN t4
>   (    t6) = parseLBRACE t5
>   (e,  t7) = parseE      t6 
>   (bv, t8) = parseBV     t7
>   (    t9) = parseRBRACE t8
>  in
>   (PROCp x xv e bv posn, t9)

	B ::= X ":" B'

> parseB :: Tokens -> (Bp, Tokens)
> parseB t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseB t0 =
>  let
>   (x, t1) = parseX     t0
>   (   t2) = parseCOLON t1
>   (b, t3) = parseB' x  t2
>  in
>   (b, t3)

	B' ::= "{" E BV "}"
	B' ::= E

> parseB' :: Xp -> Tokens -> (Bp, Tokens)
> parseB' x t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseB' x (LBRACE_TOKEN _:t1) =
>  let
>   (e,  t2) = parseE      t1
>   (bv, t3) = parseBV     t2
>   (    t4) = parseRBRACE t3
>  in
>   (GROUPp x e bv, t4)
> parseB' x t0 =
>  let
>   (e, t1) = parseE t0
>  in
>   (BLOCKp x e, t1)

	E ::= "goto" X
	E ::= "ret" V E'
	E ::= "if" V "then" E "else" E
	E ::= XV "=" E''

> parseE :: Tokens -> (Ep, Tokens)
> parseE t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseE (GOTO_TOKEN posn:t1) =
>  let
>   (x, t2) = parseX t1
>  in
>   (GOTOp x posn, t2)
> parseE (RET_TOKEN posn:t1) =
>  let
>   (v, t2) = parseV         t1
>   (e, t3) = parseE' posn v t2
>  in
>   (e, t3)
> parseE (IF_TOKEN posn:t1) =
>  let
>   (v,  t2) = parseV     t1
>   (    t3) = parseTHEN  t2
>   (e1, t4) = parseE     t3
>   (    t5) = parseELSE  t4
>   (e2, t6) = parseE     t5
>  in
>   (IFp v e1 e2 posn, t6)
> parseE t0 =
>  let
>   (xv, t1) = parseXV      t0
>   (    t2) = parseASSIGN  t1
>   (e,  t3) = parseE'' xv  t2
>  in
>   (e, t3)

	E' ::= "(" VVopt ")"
	E' ::= "," VV
	E' ::= epsilon

> parseE' :: Posn -> Vp -> Tokens -> (Ep, Tokens)
> parseE' posn v (LPAREN_TOKEN _:t1) =
>  let
>   (vv, t2) = parseVVopt  t1
>   (    t3) = parseRPAREN t2
>  in
>   (RETAp v vv posn, t3)
> parseE' posn v (COMMA_TOKEN _:t1) =
>  let
>   (vv, t2) = parseVV t1
>  in
>   (RETp (v:vv) posn, t2)
> parseE' posn v t0 = (RETp [v] posn, t0)

	E'' ::= "phi" "(" GV ")" E
	E'' ::= V E'''

> parseE'' :: XVp -> Tokens -> (Ep, Tokens)
> parseE'' xv t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseE'' [x] (PHI_TOKEN _:t1) =
>  let
>   (    t2) = parseLPAREN t1
>   (gv, t3) = parseGV     t2
>   (    t4) = parseRPAREN t3
>   (e,  t5) = parseE      t4
>  in
>   (PHIp x gv e, t5)
> parseE'' (x:xv) (PHI_TOKEN _:t1) = (undefined, [ERR_TOKEN (exprPosn x) msg])
>  where msg = "cannot define multiple variables in a \"phi\" binding"
> parseE'' xv t0 =
>  let
>   (v, t1) = parseV         t0
>   (e, t2) = parseE''' xv v t1
>  in
>   (e, t2)

	E''' ::= "(" VVopt ")" E
	E''' ::= E

> parseE''' :: XVp -> Vp -> Tokens -> (Ep, Tokens)
> parseE''' xv v t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseE''' xv v (LPAREN_TOKEN _:t1) =
>  let
>   (vv, t2) = parseVVopt  t1
>   (    t3) = parseRPAREN t2
>   (e,  t4) = parseE      t3
>  in
>   (LETAp xv v vv e, t4)
> parseE''' [x] v t0 =
>  let
>   (e, t1) = parseE t0
>  in
>   (LETp x v e, t1)
> parseE''' (x:xv) v t0 = (undefined, [ERR_TOKEN (exprPosn x) msg])
>  where msg = "cannot define multiple variables in a \"phi\" binding"

	FV ::= F FV			-- when followed by "proc"
	FV ::= epsilon

> parseFV :: Tokens -> (FVp, Tokens)
> parseFV t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseFV t0@(PROC_TOKEN _:_) =
>  let
>   (f,  t1) = parseF  t0
>   (fv, t2) = parseFV t1
>  in
>   (f:fv, t2)
> parseFV t0 = ([], t0)

	BV ::= epsilon			-- when followed by "}"
	BV ::= B BV

> parseBV :: Tokens -> (BVp, Tokens)
> parseBV t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseBV t0@(RBRACE_TOKEN _:_) = ([], t0)
> parseBV t0 =
>  let
>   (b,  t1) = parseB  t0
>   (bv, t2) = parseBV t1
>  in
>   (b:bv, t2)

	GV ::= epsilon			-- when followed by ")"
	GV ::= L ":" V GV'

> parseGV :: Tokens -> (GVp, Tokens)
> parseGV t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseGV t0@(RPAREN_TOKEN _:_) = ([], t0)
> parseGV t0 =
>  let
>   (l,  t1) = parseL     t0
>   (    t2) = parseCOLON t1
>   (v,  t3) = parseV     t2
>   (gv, t4) = parseGV'   t3
>  in
>   (FROMp l v:gv, t4)

	GV' ::= "," GV
	GV' ::= epsilon

> parseGV' :: Tokens -> (GVp, Tokens)
> parseGV' (COMMA_TOKEN _:t1) = parseGV t1
> parseGV' t0 = ([], t0)

	VVopt ::= epsilon		-- when followed by ")"
	VVopt ::= VV

> parseVVopt :: Tokens -> (VVp, Tokens)
> parseVVopt t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseVVopt t0@(RPAREN_TOKEN _:_) = ([], t0)
> parseVVopt t0 = parseVV t0

	VV ::= V VV'

> parseVV :: Tokens -> (VVp, Tokens)
> parseVV t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseVV t0 =
>  let
>   (v,  t1) = parseV   t0
>   (vv, t2) = parseVV' t1
>  in
>   (v:vv, t2)

	VV' ::= "," VV
	VV' ::= epsilon

> parseVV' :: Tokens -> (VVp, Tokens)
> parseVV' (COMMA_TOKEN _:t1) = parseVV t1
> parseVV' t0 = ([], t0)

	XVopt ::= epsilon		-- when followed by ")"
	XVopt ::= XV

> parseXVopt :: Tokens -> (XVp, Tokens)
> parseXVopt t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseXVopt t0@(RPAREN_TOKEN _:_) = ([], t0)
> parseXVopt t0 = parseXV t0

	XV ::= X XV'

> parseXV :: Tokens -> (XVp, Tokens)
> parseXV t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseXV t0@(RPAREN_TOKEN _:_) = ([], t0)
> parseXV t0 =
>  let
>   (x,  t1) = parseX   t0
>   (xv, t2) = parseXV' t1
>  in
>   (x:xv, t2)

	XV' ::= "," XV
	XV' ::= epsilon

> parseXV' :: Tokens -> (XVp, Tokens)
> parseXV' (COMMA_TOKEN _:t1) = parseXV t1
> parseXV' t0 = ([], t0)

	V -> <integer>
	V -> X

> parseV :: Tokens -> (Vp, Tokens)
> parseV t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseV (INT_TOKEN posn x:t1) = (INTp x posn, t1)
> parseV t0 =
>  let
>   (x, t1) = parseX t0
>  in
>   (VARp x, t1)

	L -> "start"
	L -> X

> parseL :: Tokens -> (Lp, Tokens)
> parseL t0@(ERR_TOKEN _ _:_) = (undefined, t0)
> parseL (START_TOKEN posn:t1) = (STARTp posn, t1)
> parseL t0 =
>  let
>   (x, t1) = parseX t0
>  in
>   (LABELp x, t1)

	And some auxiliary productions for the terminals:

> parseX :: Tokens -> (Xp, Tokens)
> parseX (ID_TOKEN posn x:t1) = (IDp 0 x posn, t1)
> parseX t0 = (undefined, panic t0 "missing identifier")

> parseTHEN :: Tokens -> Tokens
> parseTHEN (THEN_TOKEN _:t1) = t1
> parseTHEN t0 = panic t0 "missing keyword \"then\""

> parseELSE :: Tokens -> Tokens
> parseELSE (ELSE_TOKEN _:t1) = t1
> parseELSE t0 = panic t0 "missing keyword \"else\""

> parseLPAREN :: Tokens -> Tokens
> parseLPAREN (LPAREN_TOKEN _:t1) = t1
> parseLPAREN t0 = panic t0 "missing opening parenthese \"(\""

> parseRPAREN :: Tokens -> Tokens
> parseRPAREN (RPAREN_TOKEN _:t1) = t1
> parseRPAREN t0 = panic t0 "missing closing parenthese \")\""

> parseLBRACE :: Tokens -> Tokens
> parseLBRACE (LBRACE_TOKEN _:t1) = t1
> parseLBRACE t0 = panic t0 "missing opening brace \"{\""

> parseRBRACE :: Tokens -> Tokens
> parseRBRACE (RBRACE_TOKEN _:t1) = t1
> parseRBRACE t0 = panic t0 "missing closing brace \"}\""

> parseASSIGN :: Tokens -> Tokens
> parseASSIGN (ASSIGN_TOKEN _:t1) = t1
> parseASSIGN t0 = panic t0 "missing assignment operator \"=\""

> parseCOLON :: Tokens -> Tokens
> parseCOLON (COLON_TOKEN _:t1) = t1
> parseCOLON t0 = panic t0 "missing label operator \":\""

  "panic" is a simple error handler that propagates errors embedded in the
  token stream up the parse tree:

> panic :: Tokens -> String -> Tokens
> panic t0@(ERR_TOKEN _ _:_) _ = t0
> panic (t:_) msg = [ERR_TOKEN (toknPosn t) msg]
