
  SSA RENAMER
  ===========

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: SSARenamer.lhs,v 1.1 2003/03/31 09:11:12 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  The renamer assigns globally-unique identifiers to all variables, which
  greatly simplifies future program analysis. The original identifier
  information may be discarded when the debug information is stripped by
  "strip"; the new identifiers will be printed as "@n", where "n" is the
  globally-unique variable number.

  By convention, we number identifiers with positive numbers, reserving 0
  for free (undefined) variables and negative numbers for primitives.

  The "rename" function returns the new program AST.

  Unlike ANF, we must rename SSA procedures in two passes. First, we perform
  conventional top-down renaming following the variable scopes. In this pass,
  we leave the parameters to "PHI" nodes intact, since they must be renamed
  in the scope of the source of the jump.

  Note that, even though labels in Structured SSA are organized into a
  dominance tree resembling the traditional variable scoping scheme, they must
  still be given names that are unique within a procedure, since the parameter
  lists of "PHI" nodes refer to labels outside of the current scope. The same
  restriction is not enforced for variables. Instead, whenever a conflict is
  detected, the renamer replaces the corresponding "PHI" parameter by a
  "BAD_FROMp" which will be caught by the type-checker and flagged as a type
  error.

  As part of the first pass, we collect the variable environments at each exit
  points of each block. This mapping will be used by the second "fixup" pass,
  which renames the parameters to "PHI" functions.

> module SSARenamer (
>  rename
> ) where

> import List
> import FiniteMap
> import Posn
> import SSASyntax
> import SSAPrims

  First, we need an environment, which maps names to identifier numbers. Since
  labels live in a separate namespace from variables, we will store them in
  a separate "Lambda" map.

  We maintain the points of definition for all variables in "Gamma" to help
  with generation of useful error messages in "fixupG".

> type Gamma = FiniteMap String (Int, Posn)
> type Lambda = FiniteMap String Int

  The initial environment contains mappings for all primitives:

> prims :: Gamma
> prims = fst (foldl (enumerate) (emptyFM, -1) primNames)
>  where
>   enumerate (gamma, gen) name = (addToFM gamma name (gen, posn), gen - 1)
>   posn = startPosn "<prelude>"

  Now, we define functions for extending the environment and assigning
  new identifier numbers to variable names:

> introduceF :: Int -> Gamma -> Fp -> (Gamma, Int)
> introduceF gen gamma (PROCp x xv e bv _) = introduceX gen gamma x

> introduceL :: Int -> Lambda -> Xp -> (Lambda, Int)
> introduceL gen lambda (IDp _ name posn) =
>  (addToFM lambda name gen, gen + 1)

> introduceX :: Int -> Gamma -> Xp -> (Gamma, Int)
> introduceX gen gamma (IDp _ name posn) =
>  (addToFM gamma name (gen, posn), gen + 1)

> introduceB :: Int -> Lambda -> Bp -> (Lambda, Int)
> introduceB gen lambda (BLOCKp x e) = introduceL gen lambda x
> introduceB gen lambda (GROUPp x e bv) =
>  let
>   (lambda', gen') = introduceL gen lambda x
>  in
>   introduceBV gen' lambda' bv

  The vector versions of the above are a little tricky. When the list contains
  duplicate definitions of an identifier, the last definition in the list will
  replace the previous ones in the resulting "gamma". This is precisely the
  effect we want, since it makes the error visible to the type checker, which
  operates on the identifier numbers rather than names.

> introduceXV :: Int -> Gamma -> XVp -> (Gamma, Int)
> introduceXV gen gamma xv = foldl op (gamma, gen) xv
>  where op (gamma, gen) x = introduceX gen gamma x

> introduceFV :: Int -> Gamma -> FVp -> (Gamma, Int)
> introduceFV gen gamma fv = foldl op (gamma, gen) fv
>  where op (gamma, gen) f = introduceF gen gamma f

> introduceBV :: Int -> Lambda -> BVp -> (Lambda, Int)
> introduceBV gen lambda bv = foldl op (lambda, gen) bv
>  where op (lambda, gen) b = introduceB gen lambda b

  To rename "phi" nodes correctly, we require one additional data structure,
  "Beta". Because the variables appearing as parameters to a "phi" node are
  taken from the scope of the source of the jump, we must delay renaming them
  until after all relevant blocks have been renamed. The environments at the
  exit points from those blocks are returned by "renameE", and stored in a
  finite mapping "Beta". Note that since a block may contain multiple exit
  points, "Beta" maps labels to a list of environments rather than a single
  environment. It is an error for a variable used in a "phi" node to be
  defined differently in different environments, which is signalled by
  renaming the "phi" parameter to "BAD_FROMp".

> type Beta = FiniteMap Lp [Gamma]

  Now we are ready to define "rename":

> rename :: Pp -> Pp
> rename p = fst (renameP 1 prims p)

> renameP :: Int -> Gamma -> Pp -> (Pp, Int)
> renameP gen gamma (PROGp fv e) =
>  let
>   (gamma',      gen1) = introduceFV gen  gamma          fv
>   (fv',         gen2) = renameFV    gen1 gamma'         fv
>   (e', gamma_e, gen3) = renameE     gen2 gamma' emptyFM e
>  in
>   (PROGp fv' e', gen3)
> renameP gen gamma p@(SYNTAX_ERROR msg) = (p, gen)

> renameF :: Int -> Gamma -> Fp -> (Fp, Int)
> renameF gen gamma (PROCp x xv e bv posn) =
>  let
>   (gamma',      gen1) = introduceXV gen  gamma                xv
>   (lambda',     gen2) = introduceBV gen1        emptyFM       bv
>   (x'               ) = renameX          gamma                x
>   (xv'              ) = renameXV         gamma'               xv
>   (e', gamma_e, gen3) = renameE     gen2 gamma' lambda'       e
>   (beta'            ) = addToFM emptyFM (STARTp posn) gamma_e
>   (bv', beta'', gen4) = renameBV    gen3 gamma' lambda' beta' bv
>   (bv''             ) = fixupBV beta'' bv'
>  in
>   (PROCp x' xv' e' bv'' posn, gen4)

> renameB :: Int -> Gamma -> Lambda -> Beta -> Bp -> (Bp, Beta, Int)
> renameB gen gamma lambda beta (BLOCKp x e) =
>  let
>   (x'               ) = renameXL            lambda x
>   (e', gamma_e, gen1) = renameE  gen  gamma lambda e
>   beta' = addToFM beta (LABELp x') gamma_e
>  in
>   (BLOCKp x' e', beta', gen1)
> renameB gen gamma lambda beta (GROUPp x e bv) =
>  let
>   (x'               ) = renameXL               lambda       x
>   (e', gamma_e, gen1) = renameE     gen  gamma lambda       e
>   (beta'            ) = addToFM beta (LABELp x') gamma_e
>   (bv', beta'', gen2) = renameBV    gen1 gamma lambda beta' bv
>  in
>   (GROUPp x' e' bv', beta'', gen2)

  Besides the renamed expression and the updated identifier generator,
  "renameE" returns the list of variable environments at all exit points
  from the block. This is used to construct the map which renames the
  variables from the "phi" functions in the second pass over the program.

> renameE :: Int -> Gamma -> Lambda -> Ep -> (Ep, [Gamma], Int)
> renameE gen gamma lambda (PHIp x gv e) =
>  let
>   (gamma',      gen1) = introduceX gen  gamma         x
>   (x'               ) = renameX         gamma'        x
>   (gv'              ) = renameGV               lambda gv
>   (e', gamma_e, gen2) = renameE    gen1 gamma' lambda e
>  in
>   (PHIp x' gv' e', gamma_e, gen2)
> renameE gen gamma lambda (LETp x v e) =
>  let
>   (gamma',      gen1) = introduceX gen  gamma         x
>   (x'               ) = renameX         gamma'        x
>   (v'               ) = renameV         gamma         v
>   (e', gamma_e, gen2) = renameE    gen1 gamma' lambda e
>  in
>   (LETp x' v' e', gamma_e, gen2)
> renameE gen gamma lambda (LETAp xv v vv e) =
>  let
>   (gamma',      gen1) = introduceXV gen  gamma         xv
>   (xv'              ) = renameXV         gamma'        xv
>   (v'               ) = renameV          gamma         v
>   (vv'              ) = renameVV         gamma         vv
>   (e', gamma_e, gen2) = renameE     gen1 gamma' lambda e
>  in
>   (LETAp xv' v' vv' e', gamma_e, gen2)
> renameE gen gamma lambda (GOTOp x posn) =
>  let
>   (x') = renameXL lambda x
>  in
>   (GOTOp x' posn, [gamma], gen)
> renameE gen gamma lambda (RETp vv posn) =
>  let
>   (vv') = renameVV gamma vv
>  in
>   (RETp vv' posn, [gamma], gen)
> renameE gen gamma lambda (RETAp v vv posn) =
>  let
>   (v' ) = renameV  gamma v
>   (vv') = renameVV gamma vv
>  in
>   (RETAp v' vv' posn, [gamma], gen)
> renameE gen gamma lambda (IFp v e1 e2 posn) =
>  let
>   (v'                 ) = renameV      gamma        v
>   (e1', gamma_e1, gen1) = renameE gen  gamma lambda e1
>   (e2', gamma_e2, gen2) = renameE gen1 gamma lambda e2
>  in
>   (IFp v' e1' e2' posn, gamma_e1 ++ gamma_e2, gen2)

  Note that "renameG" does not rename the variables in the phi parameters;
  this is delayed until after the whole group containing the block has been
  processed.

> renameG :: Lambda -> Gp -> Gp
> renameG lambda (FROMp l v) = FROMp (renameL lambda l) v
> renameG lambda (BAD_FROMp l x posns) = BAD_FROMp (renameL lambda l) x posns

> renameV :: Gamma -> Vp -> Vp
> renameV gamma (VARp x) = VARp (renameX gamma x)
> renameV gamma v@(INTp _ _) = v

> renameL :: Lambda -> Lp -> Lp
> renameL lambda (LABELp x) = LABELp (renameXL lambda x)
> renameL lambda l@(STARTp _) = l

> renameX :: Gamma -> Xp -> Xp
> renameX gamma (IDp _ name posn) =
>  case lookupFM gamma name of
>   Just (id, _) -> IDp id name posn
>   Nothing -> IDp 0 name posn

> renameXL :: Lambda -> Xp -> Xp
> renameXL lambda (IDp _ name posn) =
>  IDp (lookupWithDefaultFM lambda 0 name) name posn

> renameFV :: Int -> Gamma -> FVp -> (FVp, Int)
> renameFV gen gamma (f:fv) =
>  let
>   (f',  gen1) = renameF  gen  gamma f
>   (fv', gen2) = renameFV gen1 gamma fv
>  in
>   (f':fv', gen2)
> renameFV gen gamma [] = ([], gen)

> renameBV :: Int -> Gamma -> Lambda -> Beta -> BVp -> (BVp, Beta, Int)
> renameBV gen gamma lambda beta (b:bv) =
>  let
>   (b',  beta1, gen1) = renameB  gen  gamma lambda beta  b
>   (bv', beta2, gen2) = renameBV gen1 gamma lambda beta1 bv
>  in
>   (b':bv', beta2, gen2)
> renameBV gen gamma lambda beta [] = ([], beta, gen)

> renameGV :: Lambda -> GVp -> GVp
> renameGV lambda gv = map (renameG lambda) gv

> renameVV :: Gamma -> VVp -> VVp
> renameVV gamma vv = map (renameV gamma) vv

> renameXV :: Gamma -> XVp -> XVp
> renameXV gamma xv = map (renameX gamma) xv

  Finally, we need to define the "PHI"-parameter renaming "fixup" pass:

> fixupB :: Beta -> Bp -> Bp
> fixupB beta (BLOCKp x e) = BLOCKp x (fixupE beta e)
> fixupB beta (GROUPp x e bv) = GROUPp x (fixupE beta e) (fixupBV beta bv)

> fixupE :: Beta -> Ep -> Ep
> fixupE beta (PHIp x gv e) = PHIp x (fixupGV beta gv) (fixupE beta e)
> fixupE beta (LETp x v e) = LETp x v (fixupE beta e)
> fixupE beta (LETAp xv v vv e) = LETAp xv v vv (fixupE beta e)
> fixupE beta (IFp v e1 e2 posn) = IFp v (fixupE beta e1) (fixupE beta e2) posn
> fixupE beta e = e

> fixupG :: Beta -> Gp -> Gp
> fixupG beta (FROMp l (VARp x@(IDp _ name posn))) =
>  let
>   gamma_e = lookupWithDefaultFM beta [] l
>   xv = foldr concat [] gamma_e
>   xv' = nubBy (\ a b -> (snd a) == (snd b)) xv
>  in
>   case xv' of
>    [(id, _)] -> FROMp l (VARp (IDp id name posn))
>    [] -> FROMp l (VARp (IDp 0 name posn))
>    _ -> BAD_FROMp l x (map (snd) xv)
>  where
>   concat gamma xs =
>    case lookupFM gamma name of
>     Just x -> x : xs
>     Nothing -> xs
> fixupG beta g = g

> fixupBV :: Beta -> BVp -> BVp
> fixupBV beta = map (fixupB beta)

> fixupGV :: Beta -> GVp -> GVp
> fixupGV beta = map (fixupG beta)
