
  LEXICAL ANALYZER FOR SSA PROGRAMS
  =================================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: SSAScanner.lhs,v 1.5 2003/03/31 09:09:39 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  This module implements a lexical analyzer for the SSA programs. The grammar
  uses the following keywords:

	proc goto ret if then else phi start = , : ( ) { }
	<identifier> <integer>
	<end of file> <lexical error>

  Both line and block comments are supported. Line comments begin with a
  sequence of two or more dashes and extend until the end of the line. Block
  comments are enclosed in "{- ... -}" delimiters and can be nested as
  expected. Whitespaces, which include all ASCII control characters, are
  ignored except as token separators.

  Integers are can be specified in base 2, 8, 10 or 16 using prefixes
  "0b", "0o", "0d" and "0x", respectively. For base-10 integers, the prefix
  is optional.

  An identifier begins with a letter or one of the symbols "_", "$", "%" or
  "@", followed by zero or more letters, digits, underscores and "'"
  characters.

  Each token is annotated with its position within the source file.  Lexical
  errors are indicated as "ERR_TOKEN" in the token stream; the analyzer does
  not attempt to recover from such errors and aborts processing after the
  first error has been encountered.

  The list of tokens returned by "scanTokens" will always end in an EOF_TOKEN,
  which permits the parser to treat syntax errors at the end of the file like
  any other syntax error during parsing.

  The function "scanTokens" processes the input list using a hand-coded
  finite-state machine derived from the lexical grammar for Etude.

> module SSAScanner (
>  Token(..), Tokens, scanTokens
> ) where

> import Posn (Posn, PChar, PString, annotateString)
> import Charset (isalpha, isdigit, isspace)

> data Token =
>  PROC_TOKEN		{ toknPosn :: Posn } |
>  GOTO_TOKEN		{ toknPosn :: Posn } |
>  RET_TOKEN		{ toknPosn :: Posn } |
>  IF_TOKEN		{ toknPosn :: Posn } |
>  THEN_TOKEN		{ toknPosn :: Posn } |
>  ELSE_TOKEN		{ toknPosn :: Posn } |
>  PHI_TOKEN		{ toknPosn :: Posn } |
>  START_TOKEN		{ toknPosn :: Posn } |
>  ASSIGN_TOKEN		{ toknPosn :: Posn } |
>  COMMA_TOKEN		{ toknPosn :: Posn } |
>  COLON_TOKEN		{ toknPosn :: Posn } |
>  LPAREN_TOKEN		{ toknPosn :: Posn } |
>  RPAREN_TOKEN		{ toknPosn :: Posn } |
>  LBRACE_TOKEN		{ toknPosn :: Posn } |
>  RBRACE_TOKEN		{ toknPosn :: Posn } |
>  ID_TOKEN		{ toknPosn :: Posn, idString :: String }  |
>  INT_TOKEN		{ toknPosn :: Posn, intValue :: Integer } |
>  ERR_TOKEN		{ toknPosn :: Posn, msgString :: String } |
>  EOF_TOKEN		{ toknPosn :: Posn }
> type Tokens = [Token]

> instance Show Token where
>  showsPrec _ (PROC_TOKEN _) = \s -> "[proc]" ++ s
>  showsPrec _ (GOTO_TOKEN _) = \s -> "[goto]" ++ s
>  showsPrec _ (RET_TOKEN _) = \s -> "[ret]" ++ s
>  showsPrec _ (IF_TOKEN _) = \s -> "[if]" ++ s
>  showsPrec _ (THEN_TOKEN _) = \s -> "[then]" ++ s
>  showsPrec _ (ELSE_TOKEN _) = \s -> "[else]" ++ s
>  showsPrec _ (PHI_TOKEN _) = \s -> "[phi]" ++ s
>  showsPrec _ (START_TOKEN _) = \s -> "[start]" ++ s
>  showsPrec _ (ASSIGN_TOKEN _) = \s -> "[=]" ++ s
>  showsPrec _ (COMMA_TOKEN _) = \s -> "[,]" ++ s
>  showsPrec _ (COLON_TOKEN _) = \s -> "[:]" ++ s
>  showsPrec _ (LPAREN_TOKEN _) = \s -> "[(]" ++ s
>  showsPrec _ (RPAREN_TOKEN _) = \s -> "[)]" ++ s
>  showsPrec _ (LBRACE_TOKEN _) = \s -> "[{]" ++ s
>  showsPrec _ (RBRACE_TOKEN _) = \s -> "[}]" ++ s
>  showsPrec _ (ID_TOKEN _ x) = \s -> "[id " ++ x ++ "]" ++ s
>  showsPrec _ (INT_TOKEN _ x) = \s -> "[int " ++ show x ++ "]" ++ s
>  showsPrec _ (ERR_TOKEN _ x) = \s -> "[error: " ++ x ++ "]" ++ s
>  showsPrec _ (EOF_TOKEN _) = \s -> "[eof]" ++ s

> scanTokens :: String -> String -> Tokens
> scanTokens name content = scanToken (annotateString name content)

> scanToken :: PString -> Tokens
> scanToken (('=',p):xs) = ASSIGN_TOKEN p : scanToken xs
> scanToken ((',',p):xs) = COMMA_TOKEN p : scanToken xs
> scanToken ((':',p):xs) = COLON_TOKEN p : scanToken xs
> scanToken (('(',p):xs) = LPAREN_TOKEN p : scanToken xs
> scanToken ((')',p):xs) = RPAREN_TOKEN p : scanToken xs
> scanToken (('}',p):xs) = RBRACE_TOKEN p : scanToken xs
> scanToken (('0',p):xs) = scan0 p xs
> scanToken (('1',p):xs) = scanDec p xs 1 True
> scanToken (('2',p):xs) = scanDec p xs 2 True
> scanToken (('3',p):xs) = scanDec p xs 3 True
> scanToken (('4',p):xs) = scanDec p xs 4 True
> scanToken (('5',p):xs) = scanDec p xs 5 True
> scanToken (('6',p):xs) = scanDec p xs 6 True
> scanToken (('7',p):xs) = scanDec p xs 7 True
> scanToken (('8',p):xs) = scanDec p xs 8 True
> scanToken (('9',p):xs) = scanDec p xs 9 True
> scanToken (('p',p):xs) = scanP p xs
> scanToken (('g',p):xs) = scanG p xs
> scanToken (('r',p):xs) = scanR p xs
> scanToken (('i',p):xs) = scanI p xs
> scanToken (('t',p):xs) = scanT p xs
> scanToken (('e',p):xs) = scanE p xs
> scanToken (('s',p):xs) = scanS p xs
> scanToken (('_',p):xs) = ID_TOKEN p ('_':x) : toks
>  where (x, toks) = scanId xs
> scanToken (('@',p):xs) = ID_TOKEN p ('@':x) : toks
>  where (x, toks) = scanId xs
> scanToken (('$',p):xs) = ID_TOKEN p ('$':x) : toks
>  where (x, toks) = scanId xs
> scanToken (('%',p):xs) = ID_TOKEN p ('%':x) : toks
>  where (x, toks) = scanId xs
> scanToken (('{',p):xs) = scanLBrace p xs
> scanToken (('-',p):xs) = scanDash p xs
> scanToken (('\r',p):xs) = [EOF_TOKEN p]
> scanToken ((c,p):xs)
>  | isalpha c = ID_TOKEN p (c:x) : toks
>  | isspace c = scanToken xs
>  | otherwise = [ERR_TOKEN p ("Unexpected character " ++ show c ++ ".")]
>  where (x, toks) = scanId xs
> scanToken [] = []

> scan0 :: Posn -> PString -> Tokens
> scan0 p (('b',p'):xs) = scanBin p xs 0 False
> scan0 p (('B',p'):xs) = scanBin p xs 0 False
> scan0 p (('o',p'):xs) = scanOct p xs 0 False
> scan0 p (('O',p'):xs) = scanOct p xs 0 False
> scan0 p (('d',p'):xs) = scanDec p xs 0 False
> scan0 p (('D',p'):xs) = scanDec p xs 0 False
> scan0 p (('x',p'):xs) = scanHex p xs 0 False
> scan0 p (('X',p'):xs) = scanHex p xs 0 False
> scan0 p (('0',p'):xs) = scanDec p xs 0 True
> scan0 p (('1',p'):xs) = scanDec p xs 1 True
> scan0 p (('2',p'):xs) = scanDec p xs 2 True
> scan0 p (('3',p'):xs) = scanDec p xs 3 True
> scan0 p (('4',p'):xs) = scanDec p xs 4 True
> scan0 p (('5',p'):xs) = scanDec p xs 5 True
> scan0 p (('6',p'):xs) = scanDec p xs 6 True
> scan0 p (('7',p'):xs) = scanDec p xs 7 True
> scan0 p (('8',p'):xs) = scanDec p xs 8 True
> scan0 p (('9',p'):xs) = scanDec p xs 9 True
> scan0 p s = INT_TOKEN p 0 : scanToken s

> scanBin :: Posn -> PString -> Integer -> Bool -> Tokens
> scanBin p (('0',p'):xs) n f = scanBin p xs (n * 2 + 0) True
> scanBin p (('1',p'):xs) n f = scanBin p xs (n * 2 + 1) True
> scanBin p s n True = INT_TOKEN p n : scanToken s
> scanBin p s n False = [ERR_TOKEN p "Expecting a binary digit."]

> scanOct :: Posn -> PString -> Integer -> Bool -> Tokens
> scanOct p (('0',p'):xs) n f = scanOct p xs (n * 8 + 0) True
> scanOct p (('1',p'):xs) n f = scanOct p xs (n * 8 + 1) True
> scanOct p (('2',p'):xs) n f = scanOct p xs (n * 8 + 2) True
> scanOct p (('3',p'):xs) n f = scanOct p xs (n * 8 + 3) True
> scanOct p (('4',p'):xs) n f = scanOct p xs (n * 8 + 4) True
> scanOct p (('5',p'):xs) n f = scanOct p xs (n * 8 + 5) True
> scanOct p (('6',p'):xs) n f = scanOct p xs (n * 8 + 6) True
> scanOct p (('7',p'):xs) n f = scanOct p xs (n * 8 + 7) True
> scanOct p s n True = INT_TOKEN p n : scanToken s
> scanOct p s n False = [ERR_TOKEN p "Expecting an octal digit."]

> scanDec :: Posn -> PString -> Integer -> Bool -> Tokens
> scanDec p (('0',p'):xs) n f = scanDec p xs (n * 10 + 0) True
> scanDec p (('1',p'):xs) n f = scanDec p xs (n * 10 + 1) True
> scanDec p (('2',p'):xs) n f = scanDec p xs (n * 10 + 2) True
> scanDec p (('3',p'):xs) n f = scanDec p xs (n * 10 + 3) True
> scanDec p (('4',p'):xs) n f = scanDec p xs (n * 10 + 4) True
> scanDec p (('5',p'):xs) n f = scanDec p xs (n * 10 + 5) True
> scanDec p (('6',p'):xs) n f = scanDec p xs (n * 10 + 6) True
> scanDec p (('7',p'):xs) n f = scanDec p xs (n * 10 + 7) True
> scanDec p (('8',p'):xs) n f = scanDec p xs (n * 10 + 8) True
> scanDec p (('9',p'):xs) n f = scanDec p xs (n * 10 + 9) True
> scanDec p s n True = INT_TOKEN p n : scanToken s
> scanDec p s n False = [ERR_TOKEN p "Expecting a decimal digit."]

> scanHex :: Posn -> PString -> Integer -> Bool -> Tokens
> scanHex p (('0',p'):xs) n f = scanHex p xs (n * 16 + 0) True
> scanHex p (('1',p'):xs) n f = scanHex p xs (n * 16 + 1) True
> scanHex p (('2',p'):xs) n f = scanHex p xs (n * 16 + 2) True
> scanHex p (('3',p'):xs) n f = scanHex p xs (n * 16 + 3) True
> scanHex p (('4',p'):xs) n f = scanHex p xs (n * 16 + 4) True
> scanHex p (('5',p'):xs) n f = scanHex p xs (n * 16 + 5) True
> scanHex p (('6',p'):xs) n f = scanHex p xs (n * 16 + 6) True
> scanHex p (('7',p'):xs) n f = scanHex p xs (n * 16 + 7) True
> scanHex p (('8',p'):xs) n f = scanHex p xs (n * 16 + 8) True
> scanHex p (('9',p'):xs) n f = scanHex p xs (n * 16 + 9) True
> scanHex p (('A',p'):xs) n f = scanHex p xs (n * 16 + 10) True
> scanHex p (('B',p'):xs) n f = scanHex p xs (n * 16 + 11) True
> scanHex p (('C',p'):xs) n f = scanHex p xs (n * 16 + 12) True
> scanHex p (('D',p'):xs) n f = scanHex p xs (n * 16 + 13) True
> scanHex p (('E',p'):xs) n f = scanHex p xs (n * 16 + 14) True
> scanHex p (('F',p'):xs) n f = scanHex p xs (n * 16 + 15) True
> scanHex p (('a',p'):xs) n f = scanHex p xs (n * 16 + 10) True
> scanHex p (('b',p'):xs) n f = scanHex p xs (n * 16 + 11) True
> scanHex p (('c',p'):xs) n f = scanHex p xs (n * 16 + 12) True
> scanHex p (('d',p'):xs) n f = scanHex p xs (n * 16 + 13) True
> scanHex p (('e',p'):xs) n f = scanHex p xs (n * 16 + 14) True
> scanHex p (('f',p'):xs) n f = scanHex p xs (n * 16 + 15) True
> scanHex p s n True = INT_TOKEN p n : scanToken s
> scanHex p s n False = [ERR_TOKEN p "Expecting a hexadecimal digit."]

> scanP :: Posn -> PString -> Tokens
> scanP p (('r',p'):xs) = scanPR p xs
> scanP p (('h',p'):xs) = scanPH p xs
> scanP p ((c,p'):xs) | isident c = ID_TOKEN p ('p':c:x) : toks
>  where (x, toks) = scanId xs
> scanP p s = ID_TOKEN p "p" : scanToken s

> scanPR :: Posn -> PString -> Tokens
> scanPR p (('o',p'):xs) = scanPRO p xs
> scanPR p ((c,p'):xs) | isident c = ID_TOKEN p ('p':'r':c:x) : toks
>  where (x, toks) = scanId xs
> scanPR p s = ID_TOKEN p "pr" : scanToken s

> scanPRO :: Posn -> PString -> Tokens
> scanPRO p (('c',p'):xs) = scanPROC p xs
> scanPRO p ((c,p'):xs) | isident c = ID_TOKEN p ('p':'r':'o':c:x) : toks
>  where (x, toks) = scanId xs
> scanPRO p s = ID_TOKEN p "pro" : scanToken s

> scanPROC :: Posn -> PString -> Tokens
> scanPROC p ((c,p'):xs) | isident c = ID_TOKEN p ('p':'r':'o':'c':c:x) : toks
>  where (x, toks) = scanId xs
> scanPROC p s = PROC_TOKEN p : scanToken s

> scanPH :: Posn -> PString -> Tokens
> scanPH p (('i',p'):xs) = scanPHI p xs
> scanPH p ((c,p'):xs) | isident c = ID_TOKEN p ('p':'h':c:x) : toks
>  where (x, toks) = scanId xs
> scanPH p s = ID_TOKEN p "ph" : scanToken s

> scanPHI :: Posn -> PString -> Tokens
> scanPHI p ((c,p'):xs) | isident c = ID_TOKEN p ('p':'h':'i':c:x) : toks
>  where (x, toks) = scanId xs
> scanPHI p s = PHI_TOKEN p : scanToken s

> scanG :: Posn -> PString -> Tokens
> scanG p (('o',p'):xs) = scanGO p xs
> scanG p ((c,p'):xs) | isident c = ID_TOKEN p ('g':c:x) : toks
>  where (x, toks) = scanId xs
> scanG p s = ID_TOKEN p "g" : scanToken s

> scanGO :: Posn -> PString -> Tokens
> scanGO p (('t',p'):xs) = scanGOT p xs
> scanGO p ((c,p'):xs) | isident c = ID_TOKEN p ('g':'o':c:x) : toks
>  where (x, toks) = scanId xs
> scanGO p s = ID_TOKEN p "go" : scanToken s

> scanGOT :: Posn -> PString -> Tokens
> scanGOT p (('o',p'):xs) = scanGOTO p xs
> scanGOT p ((c,p'):xs) | isident c = ID_TOKEN p ('g':'o':'t':c:x) : toks
>  where (x, toks) = scanId xs
> scanGOT p s = ID_TOKEN p "got" : scanToken s

> scanGOTO :: Posn -> PString -> Tokens
> scanGOTO p ((c,p'):xs) | isident c = ID_TOKEN p ('g':'o':'t':'o':c:x) : toks
>  where (x, toks) = scanId xs
> scanGOTO p s = GOTO_TOKEN p : scanToken s

> scanR :: Posn -> PString -> Tokens
> scanR p (('e',p'):xs) = scanRE p xs
> scanR p ((c,p'):xs) | isident c = ID_TOKEN p ('r':c:x) : toks
>  where (x, toks) = scanId xs
> scanR p s = ID_TOKEN p "r" : scanToken s

> scanRE :: Posn -> PString -> Tokens
> scanRE p (('t',p'):xs) = scanRET p xs
> scanRE p ((c,p'):xs) | isident c = ID_TOKEN p ('r':'e':c:x) : toks
>  where (x, toks) = scanId xs
> scanRE p s = ID_TOKEN p "re" : scanToken s

> scanRET :: Posn -> PString -> Tokens
> scanRET p ((c,p'):xs) | isident c = ID_TOKEN p ('r':'e':'t':c:x) : toks
>  where (x, toks) = scanId xs
> scanRET p s = RET_TOKEN p : scanToken s

> scanI :: Posn -> PString -> Tokens
> scanI p (('f',p'):xs) = scanIF p xs
> scanI p ((c,p'):xs) | isident c = ID_TOKEN p ('i':c:x) : toks
>  where (x, toks) = scanId xs
> scanI p s = ID_TOKEN p "i" : scanToken s

> scanIF :: Posn -> PString -> Tokens
> scanIF p ((c,p'):xs) | isident c = ID_TOKEN p ('i':'f':c:x) : toks
>  where (x, toks) = scanId xs
> scanIF p s = IF_TOKEN p : scanToken s

> scanT :: Posn -> PString -> Tokens
> scanT p (('h',p'):xs) = scanTH p xs
> scanT p ((c,p'):xs) | isident c = ID_TOKEN p ('t':c:x) : toks
>  where (x, toks) = scanId xs
> scanT p s = ID_TOKEN p  "t": scanToken s

> scanTH :: Posn -> PString -> Tokens
> scanTH p (('e',p'):xs) = scanTHE p xs
> scanTH p ((c,p'):xs) | isident c = ID_TOKEN p ('t':'h':c:x) : toks
>  where (x, toks) = scanId xs
> scanTH p s = ID_TOKEN p "th" : scanToken s

> scanTHE :: Posn -> PString -> Tokens
> scanTHE p (('n',p'):xs) = scanTHEN p xs
> scanTHE p ((c,p'):xs) | isident c = ID_TOKEN p ('t':'h':'e':c:x) : toks
>  where (x, toks) = scanId xs
> scanTHE p s = ID_TOKEN p "the" : scanToken s

> scanTHEN :: Posn -> PString -> Tokens
> scanTHEN p ((c,p'):xs) | isident c = ID_TOKEN p ('t':'h':'e':'n':c:x) : toks
>  where (x, toks) = scanId xs
> scanTHEN p s = THEN_TOKEN p : scanToken s

> scanE :: Posn -> PString -> Tokens
> scanE p (('l',p'):xs) = scanEL p xs
> scanE p ((c,p'):xs) | isident c = ID_TOKEN p ('e':c:x) : toks
>  where (x, toks) = scanId xs
> scanE p s = ID_TOKEN p "e" : scanToken s

> scanEL :: Posn -> PString -> Tokens
> scanEL p (('s',p'):xs) = scanELS p xs
> scanEL p ((c,p'):xs) | isident c = ID_TOKEN p ('e':'l':c:x) : toks
>  where (x, toks) = scanId xs
> scanEL p s = ID_TOKEN p "el" : scanToken s

> scanELS :: Posn -> PString -> Tokens
> scanELS p (('e',p'):xs) = scanELSE p xs
> scanELS p ((c,p'):xs) | isident c = ID_TOKEN p ('e':'l':'s':c:x) : toks
>  where (x, toks) = scanId xs
> scanELS p s = ID_TOKEN p "els" : scanToken s

> scanELSE :: Posn -> PString -> Tokens
> scanELSE p ((c,p'):xs) | isident c = ID_TOKEN p ('e':'l':'s':'e':c:x) : toks
>  where (x, toks) = scanId xs
> scanELSE p s = ELSE_TOKEN p : scanToken s

> scanS :: Posn -> PString -> Tokens
> scanS p (('t',p'):xs) = scanST p xs
> scanS p ((c,p'):xs) | isident c = ID_TOKEN p ('s':c:x) : toks
>  where (x, toks) = scanId xs
> scanS p s = ID_TOKEN p "s" : scanToken s

> scanST :: Posn -> PString -> Tokens
> scanST p (('a',p'):xs) = scanSTA p xs
> scanST p ((c,p'):xs) | isident c = ID_TOKEN p ('s':'t':c:x) : toks
>  where (x, toks) = scanId xs
> scanST p s = ID_TOKEN p "st" : scanToken s

> scanSTA :: Posn -> PString -> Tokens
> scanSTA p (('r',p'):xs) = scanSTAR p xs
> scanSTA p ((c,p'):xs) | isident c = ID_TOKEN p ('s':'t':'a':c:x) : toks
>  where (x, toks) = scanId xs
> scanSTA p s = ID_TOKEN p "sta" : scanToken s

> scanSTAR :: Posn -> PString -> Tokens
> scanSTAR p (('t',p'):xs) = scanSTART p xs
> scanSTAR p ((c,p'):xs) | isident c = ID_TOKEN p ('s':'t':'a':'r':c:x) : toks
>  where (x, toks) = scanId xs
> scanSTAR p s = ID_TOKEN p "star" : scanToken s

> scanSTART :: Posn -> PString -> Tokens
> scanSTART p ((c,p'):xs) | isident c =
>  ID_TOKEN p ('s':'t':'a':'r':'t':c:x) : toks
>  where (x, toks) = scanId xs
> scanSTART p s = START_TOKEN p : scanToken s

> scanId :: PString -> (String, Tokens)
> scanId ((c,p):xs) | isident c = (c:x, toks)
>  where (x, toks) = scanId xs
> scanId s = ("", scanToken s)

> scanLBrace :: Posn -> PString -> Tokens
> scanLBrace start (('-',p):xs) =
>  case skipComment start xs of
>   CONTINUE xs' -> scanToken xs'
>   ABORT err -> [err]
> scanLBrace p s = LBRACE_TOKEN p : scanToken s

> data Comment = CONTINUE PString | ABORT Token

> skipComment :: Posn -> PString -> Comment
> skipComment start (('{',p):xs) = skipLBrace start p xs
> skipComment start (('-',p):xs) = skipDash start xs
> skipComment start (_:xs) = skipComment start xs
> skipComment start []  = ABORT (ERR_TOKEN start ("Unterminated comment."))

> skipLBrace :: Posn -> Posn -> PString -> Comment
> skipLBrace start p (('-',p'):xs) =
>  case skipComment p xs of
>   CONTINUE xs' -> skipComment start xs'
>   err -> err
> skipLBrace start p (('{',p'):xs) = skipLBrace start p xs
> skipLBrace start p (_:xs) = skipComment start xs
> skipLBrace start p []  = ABORT (ERR_TOKEN start ("Unterminated comment."))

> skipDash :: Posn -> PString -> Comment
> skipDash start (('}',p):xs) = CONTINUE xs
> skipDash start (('{',p):xs) = skipLBrace start p xs
> skipDash start (('-',p):xs) = skipDash start xs
> skipDash start (_:xs) = skipComment start xs
> skipDash start []  = ABORT (ERR_TOKEN start ("Unterminated comment."))

> scanDash :: Posn -> PString -> Tokens
> scanDash p (('-',p'):xs) = skipLine xs
> scanDash p s = [ERR_TOKEN p ("Unexpected character '-'.")]

> skipLine :: PString -> Tokens
> skipLine (('\n',p'):xs) = scanToken xs
> skipLine (_:xs) = skipLine xs

> isident c = (isalpha c || isdigit c || c == '_' || c == '\'')
