
  ABSTRACT SYNTAX FOR SSA PROGRAMS
  ================================

  Copyright  2000 Patryk Zadarnowski pat@jantar.org

  Redistribution  and  use  in  source  and  binary  forms,  with  or  without
  modification, are permitted provided  that the following conditions are met:
  (1) Redistributions of  source code must retain the  above copyright notice,
  this list  of conditions and the following  disclaimer.  (2) Redistributions
  in  binary form  must reproduce  the above  copyright notice,  this  list of
  conditions and  the following disclaimer  in the documentation  and/or other
  materials provided  with the distribution.  (3)  The name of  any author may
  not  be used  to  endorse or  promote  products derived  from this  software
  without their specific prior written permission.

  THIS SOFTWARE IS  PROVIDED ``AS IS AND ANY  EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR  A PARTICULAR  PURPOSE, ARE DISCLAIMED.   IN NO EVENT  SHALL THE
  AUTHORS BE LIABLE FOR  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  OR  CONSEQUENTIAL DAMAGES  (INCLUDING, BUT  NOT LIMITED  TO,  PROCUREMENT OF
  SUBSTITUTE  GOODS OR SERVICES;  LOSS OF  USE, DATA  OR PROFITS;  OR BUSINESS
  INTERRUPTION)  HOWEVER CAUSED  AND ON  ANY THEORY  OF LIABILITY,  WHETHER IN
  CONTRACT,  STRICT LIABILITY,  OR  TORT (INCLUDING  NEGLIGENCE OR  OTHERWISE)
  ARISING IN ANY WAY  OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

	$Id: SSASyntax.lhs,v 1.6 2003/03/31 09:09:39 patrykz Exp $

   --------------------------------------------------------------------------


  OVERVIEW
  ========

  This module defines the abstract syntax for SSA programs as a set of Haskell
  types. Initially, all syntactic elements are annotated with their position
  within the input stream ("Pp" and friends.) Once the program has been
  type-checked, the position information may be discarded, and, since
  processing the annonated expressions is somewhat annoying, we define the
  second set of data structures without position annotations ("P" and
  friends.) The module also defines a pretty-printer for unannotated SSA
  expressions by instantiating the "Show" class for all AST types.

> module SSASyntax (
>  P(..), F(..), B(..), E(..), G(..), V(..), L(..), X(..),
>	FV, BV, GV, VV, XV,
>  Pp(..), Fp(..), Bp(..), Ep(..), Gp(..), Vp(..), Lp(..), Xp(..),
>	FVp, BVp, GVp, VVp, XVp,
>  SSAElem(..), strip
> ) where

> import Posn
> import Diag
> import SSAPrims

  The AST for SSA programs:

> data P  = PROG FV E
> data F  = PROC X XV E BV
> data B  = BLOCK { blockLabel :: X, blockExpr :: E } |
>	    GROUP { blockLabel :: X, blockExpr :: E, groupBlocks :: BV }
> data E  = PHI X GV E |
>           LET X V E |
>           LETA XV V VV E |
>	    GOTO X |
>	    RET VV |
>	    RETA V VV |
>	    IF V E E
> data G  = FROM L V
> data V  = VAR X | INT Integer
> data L  = LABEL X | START
> data X  = ID Int
> type FV = [F]
> type BV = [B]
> type GV = [G]
> type VV = [V]
> type XV = [X]

> instance Eq L where
>  (==) (LABEL x) (LABEL y) = (==) x y
>  (==) START START = True
>  (==) l1 l2 = False

> instance Eq X where
>  (==) (ID x) (ID y) = (==) x y

> instance Ord L where
>  compare (LABEL x) (LABEL y) = compare x y
>  compare START (LABEL _) = LT
>  compare (LABEL _) START = GT
>  compare START START = EQ

> instance Ord X where
>  compare (ID x) (ID y) = compare x y

  The AST annotated with position information:

> data Pp  = PROGp FVp Ep | SYNTAX_ERROR Msgs
> data Fp  = PROCp Xp XVp Ep BVp Posn
> data Bp  = BLOCKp { blockLabel_p :: Xp, blockExpr_p :: Ep } |
>	     GROUPp { blockLabel_p :: Xp, blockExpr_p :: Ep,
>	              groupBlocks_p :: BVp }
> data Ep  = PHIp Xp GVp Ep |
>            LETp Xp Vp Ep |
>            LETAp XVp Vp VVp Ep |
>	     GOTOp Xp Posn |
>	     RETp VVp Posn |
>	     RETAp Vp VVp Posn |
>	     IFp Vp Ep Ep Posn
> data Gp  = FROMp Lp Vp | BAD_FROMp Lp Xp [Posn]
> data Vp  = VARp Xp | INTp Integer Posn
> data Lp  = LABELp Xp | STARTp Posn
> data Xp  = IDp Int String Posn
> type FVp = [Fp]
> type BVp = [Bp]
> type GVp = [Gp]
> type VVp = [Vp]
> type XVp = [Xp]

> instance Eq Lp where
>  (==) (LABELp x) (LABELp y) = (==) x y
>  (==) (STARTp _) (STARTp _) = True
>  (==) l1 l2 = False

> instance Eq Xp where
>  (==) (IDp x _ _) (IDp y _ _) = (==) x y

> instance Ord Lp where
>  compare (LABELp x) (LABELp y) = compare x y
>  compare (STARTp _) (LABELp _) = LT
>  compare (LABELp _) (STARTp _) = GT
>  compare (STARTp _) (STARTp _) = EQ

> instance Ord Xp where
>  compare (IDp x _ _) (IDp y _ _) = compare x y

  Convinient access to the position information:

> class SSAElem a where
>  exprPosn :: a -> Posn

> instance SSAElem Pp where
>  exprPosn (PROGp (f:fv) e) = exprPosn f
>  exprPosn (PROGp [] e) = exprPosn e
>  exprPosn (SYNTAX_ERROR msgs) = undefined
> instance SSAElem Fp where
>  exprPosn (PROCp x xv e bv posn) = posn
> instance SSAElem Bp where
>  exprPosn (BLOCKp x e) = exprPosn x
>  exprPosn (GROUPp x e bv) = exprPosn x
> instance SSAElem Ep where
>  exprPosn (PHIp x gv e) = exprPosn x
>  exprPosn (LETp x v e) = exprPosn x
>  exprPosn (LETAp (x:xv) v vv e) = exprPosn x
>  exprPosn (GOTOp x posn) = posn
>  exprPosn (RETp vv posn) = posn
>  exprPosn (RETAp v vv posn) = posn
>  exprPosn (IFp x e1 e2 posn) = posn
> instance SSAElem Gp where
>  exprPosn (FROMp l v) = exprPosn l
>  exprPosn (BAD_FROMp l x posns) = exprPosn x
> instance SSAElem Vp where
>  exprPosn (VARp x) = exprPosn x
>  exprPosn (INTp x posn) = posn
> instance SSAElem Lp where
>  exprPosn (LABELp x) = exprPosn x
>  exprPosn (STARTp posn) = posn
> instance SSAElem Xp where
>  exprPosn (IDp x name posn) = posn

  A function which removes position informatoin from the program:

> strip :: Pp -> P
> strip = stripP

> stripP :: Pp -> P
> stripP (PROGp fv e) = PROG (stripFV fv) (stripE e)
> stripP (SYNTAX_ERROR msgs) = error "cannot strip an invalid program"

> stripF :: Fp -> F
> stripF (PROCp x xv e bv _) =
>  PROC (stripX x) (stripXV xv) (stripE e) (stripBV bv)

> stripB :: Bp -> B
> stripB (BLOCKp x e) = BLOCK (stripX x) (stripE e)
> stripB (GROUPp x e bv) = GROUP (stripX x) (stripE e) (stripBV bv)

> stripE :: Ep -> E
> stripE (PHIp x gv e) = PHI (stripX x) (stripGV gv) (stripE e)
> stripE (LETp x v e) = LET (stripX x) (stripV v) (stripE e)
> stripE (LETAp xv v vv e) =
>  LETA (stripXV xv) (stripV v) (stripVV vv) (stripE e)
> stripE (GOTOp x _) = GOTO (stripX x)
> stripE (RETp vv _) = RET (stripVV vv)
> stripE (RETAp v vv _) = RETA (stripV v) (stripVV vv)
> stripE (IFp v e1 e2 _) = IF (stripV v) (stripE e1) (stripE e2)

> stripG :: Gp -> G
> stripG (FROMp l v) = FROM (stripL l) (stripV v)
> stripG (BAD_FROMp l x posns) = error "cannot strip an illegal program"

> stripV :: Vp -> V
> stripV (VARp x) = VAR (stripX x)
> stripV (INTp x _) = INT x

> stripL :: Lp -> L
> stripL (LABELp x) = LABEL (stripX x)
> stripL (STARTp _) = START

> stripX :: Xp -> X
> stripX (IDp x _ _) = ID x

> stripFV :: FVp -> FV
> stripFV = map (stripF)

> stripBV :: BVp -> BV
> stripBV = map (stripB)

> stripGV :: GVp -> GV
> stripGV = map (stripG)

> stripVV :: VVp -> VV
> stripVV = map (stripV)

> stripXV :: XVp -> XV
> stripXV = map (stripX)

  The pretty-printer for stripped expressions:

> instance Show P where
>  showsPrec _ = showP
> instance Show F where
>  showsPrec _ = showF
>  showList = showFV
> instance Show B where
>  showsPrec _ = showB 1
>  showList = showBV 1
> instance Show E where
>  showsPrec _ = showE 0
> instance Show G where
>  showsPrec _ = showG
>  showList = showGV
> instance Show V where
>  showsPrec _ = showV
>  showList = showVV
> instance Show L where
>  showsPrec _ = showL
>  showList = showA
> instance Show X where
>  showsPrec _ = showX
>  showList = showXV

> showP :: P -> ShowS
> showP (PROG [] e) =
>  showE 0 e
> showP (PROG fv e) =
>  showFV fv . showC '\n' .
>  showC '\n' .
>  showE 0 e

> showF :: F -> ShowS
> showF (PROC x xv e bv) =
>  showS "proc " . showX x . showC '(' . showXV xv . showS ") {\n" .
>  showC '\n' .
>  showC '\t' . showEBV 1 e bv . showC '\n' .
>  showC '\n' .
>  showC '}'

> showB :: Int -> B -> ShowS
> showB i (BLOCK x e) =
>  let
>   n = idLength x
>   m = i * 8 - 2
>  in
>   if n <= m then
>    indent ' ' (m - n) . showX x . showS ":\t" . showE i e
>   else
>    showX x . showS ":\n" .
>    showC '\n' .
>    indent '\t' i . showE i e
> showB i (GROUP x e bv) =
>  let
>   n = idLength x
>   m = i * 8 - 2
>  in
>   if n <= m then
>    indent ' ' (m - n) . showX x . showS ":\t{\n"
>   else
>    showX x . showS ": {\n" .
>   showC '\n' .
>   indent '\t' (i + 1) . showEBV (i + 1) e bv . showC '\n' .
>   showC '\n' .
>   indent '\t' i . showC '}'

> showE :: Int -> E -> ShowS
> showE i (PHI x gv e) =
>  showX x . showS " = phi(" . showGV gv . showS ")\n" .
>  indent '\t' i . showE i e
> showE i (LET x v e) =
>  showX x . showS " = " . showV v . showC '\n' .
>  indent '\t' i . showE i e
> showE i (LETA xv v vv e) =
>  showXV xv . showS " = " . showV v . showC '(' . showVV vv . showS ")\n" .
>  indent '\t' i . showE i e
> showE i (GOTO x) = showS "goto " . showX x
> showE i (RET vv) = showS "ret " . showVV vv
> showE i (RETA v vv) =
>  showS "ret " . showV v . showC '(' . showVV vv . showC ')'
> showE i (IF v e1 e2) =
>  showS "if " . showV v . showS " then\n" .
>  indent '\t' (i + 1) . showE (i + 1) e1 . showC '\n' .
>  indent '\t' i . showS "else\n" .
>  indent '\t' (i + 1) . showE (i + 1) e2

> showEBV :: Int -> E -> BV -> ShowS
> showEBV i e [] =
>  showE i e
> showEBV i e bv =
>  showE i e . showC '\n' .
>  showC '\n' .
>  showBV i bv

> showG :: G -> ShowS
> showG (FROM l v) = showL l . showC ':' . showV v

> showV :: V -> ShowS
> showV (VAR x) = showX x
> showV (INT x) = shows x

> showL :: L -> ShowS
> showL (LABEL x) = showX x
> showL (START) = showS "start"

> showX :: X -> ShowS
> showX (ID id) =
>  if id < 0 then
>   if -id <= length primNames then
>    showS (primNames!!(-id - 1))
>   else
>    showS "@@" . shows (-id)
>  else
>   showC '@' . shows id

> showFV :: FV -> ShowS
> showFV = showQ (showF)

> showBV :: Int -> BV -> ShowS
> showBV i = showQ (showB i)

> showGV :: GV -> ShowS
> showGV = showA

> showVV :: VV -> ShowS
> showVV = showA

> showXV :: XV -> ShowS
> showXV = showA

> idLength :: X -> Int
> idLength x = length (show x)

  Now the variant for annotated expression:

> instance Show Pp where
>  showsPrec _ = showPp
> instance Show Fp where
>  showsPrec _ = showFp
>  showList = showFVp
> instance Show Bp where
>  showsPrec _ = showBp 1
>  showList = showBVp 1
> instance Show Ep where
>  showsPrec _ = showEp 0
> instance Show Gp where
>  showsPrec _ = showGp
>  showList = showGVp
> instance Show Vp where
>  showsPrec _ = showVp
>  showList = showVVp
> instance Show Lp where
>  showsPrec _ = showLp
>  showList = showA
> instance Show Xp where
>  showsPrec _ = showXp
>  showList = showXVp

> showPp :: Pp -> ShowS
> showPp (PROGp [] e) =
>  showEp 0 e
> showPp (PROGp fv e) =
>  showFVp fv . showC '\n' .
>  showC '\n' .
>  showEp 0 e
> showPp (SYNTAX_ERROR msgs) = shows msgs

> showFp :: Fp -> ShowS
> showFp (PROCp x xv e bv _) =
>  showS "proc " . showXp x . showC '(' . showXVp xv . showS ") {\n" .
>  showC '\n' .
>  showC '\t' . showEBVp 1 e bv . showC '\n' .
>  showC '\n' .
>  showC '}'

> showBp :: Int -> Bp -> ShowS
> showBp i (BLOCKp x e) =
>  let
>   n = idpLength x
>   m = i * 8 - 2
>  in
>   if n <= m then
>    indent ' ' (m - n) . showXp x . showS ":\t" . showEp i e
>   else
>    showXp x . showS ":\n" .
>    showC '\n' .
>    indent '\t' i . showEp i e
> showBp i (GROUPp x e bv) =
>  let
>   n = idpLength x
>   m = i * 8 - 2
>  in
>   if n <= m then
>    indent ' ' (m - n) . showXp x . showS ":\t{\n"
>   else
>    showXp x . showS ": {\n" .
>   showC '\n' .
>   indent '\t' (i + 1) . showEBVp (i + 1) e bv . showC '\n' .
>   showC '\n' .
>   indent '\t' i . showC '}'

> showEp :: Int -> Ep -> ShowS
> showEp i (PHIp x gv e) =
>  showXp x . showS " = phi(" . showGVp gv . showS ")\n" .
>  indent '\t' i . showEp i e
> showEp i (LETp x v e) =
>  showXp x . showS " = " . showVp v . showC '\n' .
>  indent '\t' i . showEp i e
> showEp i (LETAp xv v vv e) =
>  showXVp xv . showS " = " . showVp v . showC '(' . showVVp vv . showS ")\n" .
>  indent '\t' i . showEp i e
> showEp i (GOTOp x _) = showS "goto " . showXp x
> showEp i (RETp vv _) = showS "ret " . showVVp vv
> showEp i (RETAp v vv _) =
>  showS "ret " . showVp v . showC '(' . showVVp vv . showC ')'
> showEp i (IFp v e1 e2 _) =
>  showS "if " . showVp v . showS " then\n" .
>  indent '\t' (i + 1) . showEp (i + 1) e1 . showC '\n' .
>  indent '\t' i . showS "else\n" .
>  indent '\t' (i + 1) . showEp (i + 1) e2

> showEBVp :: Int -> Ep -> BVp -> ShowS
> showEBVp i e [] =
>  showEp i e
> showEBVp i e bv =
>  showEp i e . showC '\n' .
>  showC '\n' .
>  showBVp i bv

> showGp :: Gp -> ShowS
> showGp (FROMp l v) = showLp l . showC ':' . showVp v
> showGp (BAD_FROMp l x posns) = showLp l . showC ':' . showXp x
>  . shows (map (lineno) posns)

> showVp :: Vp -> ShowS
> showVp (VARp x) = showXp x
> showVp (INTp x _) = shows x

> showLp :: Lp -> ShowS
> showLp (LABELp x) = showXp x
> showLp (STARTp _) = showS "start"

> showXp :: Xp -> ShowS
> showXp (IDp x name _) = showS name

> showFVp :: FVp -> ShowS
> showFVp = showQ (showFp)

> showBVp :: Int -> BVp -> ShowS
> showBVp i = showQ (showBp i)

> showGVp :: GVp -> ShowS
> showGVp = showA

> showVVp :: VVp -> ShowS
> showVVp = showA

> showXVp :: XVp -> ShowS
> showXVp = showA

> idpLength :: Xp -> Int
> idpLength (IDp x name _) = length name

  And some additional helpers:

> showQ :: (Show a) => (a -> ShowS) -> [a] -> ShowS
> showQ shows (a:av) = shows a . showQ' av
>  where
>   showQ' [] = id
>   showQ' av = showS "\n\n" . showQ shows av
> showQ shows [] = id

> showA :: (Show a) => [a] -> ShowS
> showA (a:av) = shows a . showA' av
>  where
>   showA' [] = id
>   showA' av = showS ", " . showA av
> showA [] = id

> showS :: String -> ShowS
> showS (c:s) = showC c . showS s
> showS [] = id

> showC :: Char -> ShowS
> showC c = \ s -> c : s

> indent :: Char -> Int -> ShowS
> indent c 0 = id
> indent c n = showC c . indent c (n - 1)
